
from __future__ import print_function, absolute_import

import os, sys, subprocess, glob
from ..coloring import Coloring

class PathMixIn( object ):
    _dev_null = None
    
    def _callExternal( self, cmd, suppress_fail=False, verbose=True ) :
        try:
            if verbose and not suppress_fail:
                print( Coloring.infoLabel('Call: ') +
                      Coloring.info(subprocess.list2cmdline( cmd )),
                      file=sys.stderr )
                stderr = sys.stderr
            else:
                if not PathMixIn._dev_null:
                    PathMixIn._dev_null = open(os.devnull, 'w')
                stderr = PathMixIn._dev_null
            res = subprocess.check_output( cmd, stdin=subprocess.PIPE, stderr=stderr )
            
            try:
                res = str(res, 'utf8')
            except:
                pass
            
            return res
        except subprocess.CalledProcessError as e:
            if suppress_fail :
                return None
            raise e
        
    def _callInteractive( self, cmd, replace=True ):
        if replace:
            print( Coloring.infoLabel('Exec: ') +
                    Coloring.info(subprocess.list2cmdline( cmd )),
                    file=sys.stderr )

            sys.stdout.flush()
            sys.stderr.flush()
            os.execv(cmd[0], cmd)
        else:
            print( Coloring.infoLabel('Call: ') +
                    Coloring.info(subprocess.list2cmdline( cmd )),
                    file=sys.stderr )
            sys.stdout.flush()
            sys.stderr.flush()
            
            return subprocess.check_call( cmd )
        
    def _trySudoCall( self, cmd, errmsg=None ):
        try:
            self._callExternal(['sudo', '-n'] + cmd)
        except subprocess.CalledProcessError:
            if not errmsg:
                errmsg = 'you may need to call the the failed command manually !'

            self._warn(errmsg)
    
    def _which( self, program ):
        "Copied from stackoverflow"
        def is_exe(fpath):
            return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

        fpath, fname = os.path.split(program)
        if fpath:
            if is_exe(program):
                return program
        else:
            for path in os.environ["PATH"].split(os.pathsep):
                path = path.strip('"')
                exe_file = os.path.join(path, program)
                if is_exe(exe_file):
                    return exe_file

        return None

    def _addEnvPath( self, env_name, add_dir, first=False ) :
        if env_name in os.environ:
            dir_list = os.environ[env_name].split(os.pathsep)
        else :
            dir_list = []

        if add_dir not in dir_list:
            if first :
                dir_list[0:0] = [add_dir]
            else :
                dir_list.append(add_dir)

            os.environ[env_name] = os.pathsep.join(dir_list)

    def _addBinPath( self, bin_dir, first=False ) :
        self._addEnvPath( 'PATH', bin_dir, first=first)

    def _addPackageFiles( self, config, pattern ):
        files = glob.glob(pattern)
        
        if not files:
            self._errorExit('Failed to find created packages of "{0}" pattern'.format(pattern))
        
        config.setdefault('packageFiles', [])
        config['packageFiles'] += files
        



from ..runenvtool import RunEnvTool

class curlTool( RunEnvTool ):
    """Command line tool and library for transferring data with URLs.

Home: https://curl.haxx.se/
"""
    def _installTool( self, env ):
        self._requirePackages(['curl'])
        self._requireEmerge(['net-misc/curl'])
        self._requirePacman(['curl'])
        self._requireHomebrew('curl')




from __future__ import print_function

import os

from ..runtimetool import RuntimeTool

class pythonTool( RuntimeTool ):
    """Python is a programming language.
    
Home: https://www.python.org/

Only the first part of pythonVer is used for installation of
system packages OS-specific way.
"""    
    VER_CMD = 'import sys; print( "%s.%s" % (sys.version_info.major, sys.version_info.minor) )'
    
    def getPostDeps( self ) :
        # hackish hack
        return ['virtualenv']
    
    def envNames( self ) :
        return ['pythonBin', 'pythonVer']
    
    def _installTool( self, env ):
        if int(env['pythonVer'].split('.')[0]) == 3:
            self._requireDeb(['python3'])
            self._requireZypper(['python3'])
                
            self._requireYumEPEL()
            self._requireYum(['python34'])
            self._requirePacman(['python'])
            self._requireHomebrew('python3')
        else:
            self._requirePackages(['python'])
            self._requirePacman(['python2'])
            self._requireHomebrew('python')

        self._requireEmerge(['=dev-lang/python-{0}*'.format(env['pythonVer'])])
    
    def uninstallTool( self, env ):
        pass

    def initEnv( self, env ) :
        python_ver = env.setdefault('pythonVer', '3')
        bin_name = None
        
        if self._isGentoo():
            if len(python_ver) == 3:
                os.environ['EPYTHON'] = 'python{0}'.format(python_ver)
            elif int(python_ver.split('.')[0]) == 3:
                os.environ['EPYTHON'] = 'python3.4'
            else:
                os.environ['EPYTHON'] = 'python2.7'
        elif self._isArchLinux():
            if int(python_ver.split('.')[0]) == 2:
                bin_name = 'python2'
        elif int(python_ver.split('.')[0]) == 3:
            bin_name = 'python3'

        super(pythonTool, self).initEnv( env, bin_name )
        
        if self._have_tool and 'pythonRawBin' not in env:
            env['pythonRawBin'] = env['pythonBin']
            python_ver_fact = self._callExternal(
                [ env['pythonRawBin'], '-c', self.VER_CMD ],
                verbose = False
            ).strip()
            
            if python_ver.split('.') > python_ver_fact.split('.'):
                self._errorExit(
                    'Too old python version "{0}" when "{1}" is required'
                    .format(python_ver, python_ver_fact)
                )

            env['pythonFactVer'] = python_ver_fact




import os

from ..buildtool import BuildTool
from ..testtool import TestTool
from .piptoolmixin import PipToolMixIn

class setuptoolsTool( PipToolMixIn, BuildTool, TestTool ):
    """Easily download, build, install, upgrade, and uninstall Python packages.
    
Home: https://pypi.python.org/pypi/setuptools

Not assumed to be used directly.

Build targets:
    prepare -> {removes build & dist folders}
    build -> ['sdist', 'bdist_wheel']
    package -> {uses result of build in dist/}
Override targets with .config.toolTune.    

"""    
    def autoDetectFiles( self ) :
        return 'setup.py'
    
    def uninstallTool( self, env ):
        pass
        
    def initEnv( self, env ):
        virtualenv_dir = env['virtualenvDir']
        self._have_tool = os.path.exists(os.path.join(virtualenv_dir, 'bin', 'easy_install'))
        
    def onPrepare( self, config ):
        targets = self._getTune(config, 'prepare', ['build', 'dist'])
        
        if not isinstance(targets, list):
            targets = [targets]
        
        for d in targets:
            if os.path.exists(d):
                self._rmTree(d)

    def onBuild( self, config ):
        env = config['env']
        self._requirePip( env, 'wheel' )
        
        targets = self._getTune(config, 'build', ['sdist', 'bdist_wheel'])

        if not isinstance(targets, list):
            targets = [targets]
        
        self._callExternal( [ env['pythonBin'], 'setup.py' ] + targets )
    
    def onPackage(self, config):
        target = self._getTune(config, 'package', 'dist')
        self._addPackageFiles(config, os.path.join(target, '*'))

    def onCheck( self, config ):
        env = config['env']
        self._requirePip( env, 'docutils' )
        self._requirePip( env, 'readme' )
        self._callExternal( [ env['pythonBin'], 'setup.py', 'check', '-mrs' ] )

    



import sys

class Coloring:
    _color = sys.stdout.isatty()
    
    @classmethod
    def enable(cls, val):
        cls._color = val
    
    @classmethod
    def wrap(cls, text, escape):
        if cls._color:
            ESC = chr(0x1B)
            return ESC + escape + str(text) + ESC + '[0m'
        else:
            return text
        
    @classmethod
    def error(cls, text):
        # bright red on black
        return cls.wrap(text, '[1;31;40m')
    
    @classmethod
    def warn(cls, text):
        # yellow on black
        return cls.wrap(text, '[33;40m')
    
    @classmethod
    def info(cls, text):
        # cyan on black
        return cls.wrap(text, '[36;40m')
    
    @classmethod
    def infoLabel(cls, text):
        # bright cyan on black
        return cls.wrap(text, '[1;36;40m')

    @classmethod
    def label(cls, text):
        # bright
        return cls.wrap(text, '[1m')



#!/usr/bin/env python
"""FutoIn Continuous Integration & Delivery Tool.

Usage:
    cid init [<project_name>] [--vcsRepo=<vcs_repo>] [--rmsRepo=<rms_repo>] [--permissive]
    cid tag <branch> [<next_version>] [--vcsRepo=<vcs_repo>] [--wcDir=<wc_dir>]
    cid prepare [<vcs_ref>] [--vcsRepo=<vcs_repo>] [--wcDir=<wc_dir>]
    cid build [--debug]
    cid package
    cid check [--permissive]
    cid promote <package> <rms_pool> [--rmsRepo=<rms_repo>] [--rmsHash=<rms_hash>]
    cid deploy vcstag [<vcs_ref>] [--vcsRepo=<vcs_repo>] [--redeploy] [--deployDir=<deploy_dir>]
    cid deploy vcsref <vcs_ref> [--vcsRepo=<vcs_repo>] [--redeploy] [--deployDir=<deploy_dir>]
    cid deploy [rms] <rms_pool> [<package>] [--rmsRepo=<rms_repo>] [--rmsHash=<rms_hash>] [--redeploy] [--deployDir=<deploy_dir>] [--build]
    cid migrate
    cid run
    cid run <command> [<command_arg>...]
    cid ci_build <vcs_ref> [<rms_pool>] [--vcsRepo=<vcs_repo>] [--rmsRepo=<rms_repo>] [--permissive] [--debug] [--wcDir=<wc_dir>]
    cid tool exec <tool_name> [<tool_version>] [-- <tool_arg>...]
    cid tool (install|uninstall|update|test|env) [<tool_name> [<tool_version>]]
    cid tool (prepare|build|check|package|migrate) <tool_name> [<tool_version>]
    cid tool list
    cid tool describe <tool_name>
    cid tool detect
    cid vcs checkout [<vcs_ref>] [--vcsRepo=<vcs_repo>] [--wcDir=<wc_dir>]
    cid vcs commit <commit_msg> [<commit_files>...] [--wcDir=<wc_dir>]
    cid vcs merge <vcs_ref> [--no-cleanup] [--wcDir=<wc_dir>]
    cid vcs branch <vcs_ref> [--wcDir=<wc_dir>]
    cid vcs delete <vcs_ref> [--vcsRepo=<vcs_repo>] [--cacheDir=<cache_dir>] [--wcDir=<wc_dir>]
    cid vcs export <vcs_ref> <dst_dir> [--vcsRepo=<vcs_repo>] [--cacheDir=<cache_dir>] [--wcDir=<wc_dir>]
    cid vcs tags [<tag_pattern>] [--vcsRepo=<vcs_repo>] [--cacheDir=<cache_dir>] [--wcDir=<wc_dir>]
    cid vcs branches [<branch_pattern>] [--vcsRepo=<vcs_repo>] [--cacheDir=<cache_dir>] [--wcDir=<wc_dir>]
    cid vcs reset [--wcDir=<wc_dir>]
    cid vcs ismerged <vcs_ref> [--wcDir=<wc_dir>]
    

Options:
    -h --help                       Show this screen.
    --vcsRepo=<vcs_repo>            VCS repository URL in vcs_type:vcs_url format.
    --rmsRepo=<rms_repo>            RMS repository URL in rms_type:rms_url format.
    --wcDir=<wc_dir>                Working copy directory (project root).
    --rmsHash=<rms_hash>            Package hash for validation in hash_type:value format.
    --deployDir=<deploy_dir>        Destination for deployment.
    --redeploy                      Force redeploy.
    --build                         Build during deploy.
    --permissive                    Ignore test failures.
    --debug                         Build in debug mode, if applicable.
    --cacheDir=<cache_dir>          Directory to hold VCS cache.
"""

from __future__ import print_function, absolute_import

from .cidtool import CIDTool
from .coloring import Coloring

try:
    from docopt import docopt
except ImportError:
    # fallback to "hardcoded"
    from futoin.cid.contrib.docopt import docopt

import os, sys, traceback

if sys.version_info < (2,7):
    print( 'Sorry, but only Python version >= 2.7 is supported!', file=sys.stderr )
    sys.exit( 1 )

from . import __version__ as version
__all__ = ['run']

def runInner():
    args = docopt( __doc__, version='FutoIn CID v{0}'.format(version) )

    if type(args) == str:
        print(args)
    else:
        if 'CID_COLOR' in os.environ:
            Coloring.enable(os.environ['CID_COLOR'] == 'yes')
        
        overrides = {}
        #---
        vcsArg = args.get('--vcsRepo', None)

        if vcsArg:
            try:
                (overrides['vcs'],
                overrides['vcsRepo']) = vcsArg.split(':', 1)
            except ValueError:
                raise RuntimeError('Invalid argument to --vcsRepo')
        #---
        rmsArg = args.get('--rmsRepo', None)

        if rmsArg:
            try:
                (overrides['rms'],
                overrides['rmsRepo']) = rmsArg.split(':', 1)
            except ValueError:
                raise RuntimeError('Invalid argument to --rmsRepo')
        overrides['rmsHash'] = args['--rmsHash']
        overrides['rmsPool'] = args['<rms_pool>']
        #---
        if args['ci_build'] :
            if 'vcs' in overrides:
                def_wc_dir = 'ci_build'
            else :
                def_wc_dir = os.path.join('..', 'ci_builds',
                                          os.path.basename(os.path.realpath('.')))
            def_wc_dir += '_' + args['<vcs_ref>']
        else :
            def_wc_dir = '.'
        overrides['wcDir'] = os.path.realpath(args['--wcDir'] or def_wc_dir)
        #--
        deploy_dir = args['--deployDir']
        overrides['deployDir'] = deploy_dir and os.path.realpath(deploy_dir) or None
        overrides['reDeploy'] = args['--redeploy'] and True or False

        if args['--build']:
            overrides['deployBuild'] = True

        if args['--debug']:
            overrides['debugBuild'] = True
        
        # enable vcsref & vcstag build by default
        if args['deploy'] and (args['vcsref'] or args['vcstag']):
            overrides['deployBuild'] = True

        #---
        overrides['vcsRef'] = args['<vcs_ref>']

        #---
        tool = args['<tool_name>']
        tool_ver = args['<tool_version>']
        overrides['tool'] = tool
        overrides['toolVer'] = tool_ver != '--' and tool_ver or None
        overrides['toolTest'] = (
            args['test'] or
            args['uninstall'] or
            args['describe'] or
            args['detect']
        )
        
        #---
        if args['--permissive']:
            overrides['permissiveChecks'] = True
         
        #---
        cache_dir = args['--cacheDir']
        
        if cache_dir:
            cache_dir = os.path.realpath(cache_dir)
        
        #---
        cit = CIDTool( overrides = overrides )
        
        if args['tool'] :
            if args['exec']:
                cit.tool_exec( tool, args['<tool_arg>'] )
            elif args['list']:
                cit.tool_list()
            elif args['detect']:
                cit.tool_detect()
            else:
                subcmds = [
                    'install',
                    'uninstall',
                    'update',
                    'test',
                    'env',
                    'prepare',
                    'build',
                    'check',
                    'package',
                    'migrate',
                    'describe',
                ]
                for cmd in subcmds:
                    if args[cmd]:
                        getattr( cit, 'tool_'+cmd )( tool )
                        break
                else:
                    raise RuntimeError( "Unknown Command" )
        elif args['vcs'] :
            if args['checkout']:
                cit.vcs_checkout( args['<vcs_ref>'] )
            elif args['commit']:
                cit.vcs_commit( args['<commit_msg>'], args['<commit_files>'] )
            elif args['branch']:
                cit.vcs_branch( args['<vcs_ref>'] )
            elif args['merge']:
                cit.vcs_merge( args['<vcs_ref>'], not args['--no-cleanup'] )
            elif args['delete']:
                cit.vcs_delete( args['<vcs_ref>'], cache_dir )
            elif args['export']:
                dst_dir = os.path.realpath(args['<dst_dir>'])
                cit.vcs_export( args['<vcs_ref>'], dst_dir, cache_dir )
            elif args['tags']:
                cit.vcs_tags( args['<tag_pattern>'], cache_dir )
            elif args['branches']:
                cit.vcs_branches( args['<branch_pattern>'], cache_dir )
            elif args['reset']:
                cit.vcs_reset()
            elif args['ismerged']:
                cit.vcs_ismerged( args['<vcs_ref>'] )
            else:
                raise RuntimeError( "Not implemented yet." )
        elif args['init'] :
            cit.init_project( args['<project_name>'] )
        elif args['tag'] :
            cit.tag( args['<branch>'], args['<next_version>'] )
        elif args['prepare'] :
            cit.prepare( args['<vcs_ref>'] )
        elif args['build'] :
            cit.build()
        elif args['package']:
            cit.package()
        elif args['check'] :
            cit.check()
        elif args['promote'] :
            cit.promote( args['<package>'], args['<rms_pool>'] )
        elif args['deploy'] :
            if args['vcsref']:
                cit.deploy( 'vcsref', args['<vcs_ref>'] )
            elif args['vcstag']:
                cit.deploy( 'vcstag', args['<vcs_ref>'] )
            else :
                cit.deploy( 'rms', args['<rms_pool>'], args['<package>'] )
        elif args['migrate'] :
            cit.migrate()
        elif args['run'] :
            cit.run( args['<command>'], args['<command_arg>'] )
        elif args['ci_build'] :
            cit.ci_build( args['<vcs_ref>'], args['<rms_pool>'] )
        else:
            raise RuntimeError( "Unknown Command" )

def run():
    try:
        runInner()
    except Exception as e:
        print(file=sys.stderr)
        print(Coloring.error('ERROR: ' + str(e)), file=sys.stderr)
        print(file=sys.stderr)
        print(Coloring.warn(traceback.format_exc()), file=sys.stderr)
        sys.exit(1)
    except KeyboardInterrupt as e:
        print(file=sys.stderr)
        print(Coloring.error('Exit on user abort'), file=sys.stderr)
        print(file=sys.stderr)
        sys.exit(1)
        



import os

from ..runenvtool import RunEnvTool

class rustTool( RunEnvTool ):
    """Rust is a systems programming language.
    
Home: https://www.rust-lang.org
"""
    def getDeps( self ) :
        return [ 'rustup' ]
    
    def getVersionParts( self ):
        return 3

    def _installTool( self, env ):
        self._callExternal( [
            env['rustupBin'], 'toolchain', 'install', env['rustVer']
        ])
            
    def updateTool( self, env ):
        self._installTool( env )
        
    def uninstallTool( self, env ):
        self._callExternal( [
            env['rustupBin'], 'toolchain', 'uninstall', env['rustVer']
        ])
        self._have_tool = False

    def envNames( self ) :
        return ['rustBin', 'rustVer']
    
    def initEnv( self, env ) :
        ver = env.setdefault('rustVer', 'stable')
        os.environ['RUSTUP_TOOLCHAIN'] = ver
        
        try:
            res = self._callExternal( [
                env['rustupBin'], 'which', 'rustc'
            ], verbose=False)
        except:
            return

        super(rustTool, self).initEnv( env, 'rustc' )




import os
import subprocess
import tempfile

from ..vcstool import VcsTool
from .bashtoolmixin import BashToolMixIn

class gitTool( BashToolMixIn, VcsTool ):
    """Git distributed version control system.
    
Home: https://git-scm.com/

Git tool forcibly sets user.email and user.name, 
if not set by user.
"""    
    def getDeps(self):
        return ['bash', 'tar']

    def autoDetectFiles( self ) :
        return '.git'
    
    def _installTool( self, env ):
        self._requirePackages(['git'])
        self._requireEmerge(['dev-vcs/git'])
        self._requirePacman(['git'])
        self._requireHomebrew('git')

    def _checkGitConfig( self, env ):
        gitBin = env['gitBin']
        user_email = None
        user_name = None
        
        try:
            user_email = self._callExternal([
                gitBin, 'config', 'user.email',
            ], verbose=False).strip()
        except:
            pass
        
        try:
            user_name = self._callExternal([
                gitBin, 'config', 'user.name',
            ], verbose=False).strip()
        except:
            pass
        
        if not user_email:
            self._info('Setting fallback user.email for repo')
            self._callExternal([
                gitBin, 'config', 'user.email',
                env.get('gitUserEmail', 'noreply@futoin.org')
            ])

        if not user_name:
            self._info('Setting fallback user.name for repo')
            self._callExternal([
                gitBin, 'config', 'user.name',
                env.get('gitUserName', 'FutoIn CITool')
            ])
            
    def _getCurrentBranch( self, config ):
        return self._callExternal( [
            config['env']['gitBin'], 'rev-parse', '--abbrev-ref', 'HEAD'
        ], verbose=False ).strip()

    def vcsGetRepo( self, config, wc_dir=None ):
        git_dir = wc_dir or os.path.join(os.getcwd(), '.git')
        
        return self._callExternal( [
            config['env']['gitBin'],
            '--git-dir={0}'.format(git_dir),
            'config',
            '--get',
            'remote.origin.url'
        ], verbose=False ).strip()
    
    def _gitCompareRepo( self, cfg, act ):
        return cfg == act or ('ssh://'+cfg) == act
    
    def vcsCheckout( self, config, vcs_ref ):
        gitBin = config['env']['gitBin']
        wc_dir = os.getcwd()
        vcsRepo = config['vcsRepo']
        vcs_ref = vcs_ref or 'master'
        
        if os.path.isdir( '.git' ):
            remote_url = self.vcsGetRepo( config, '.git')
            
            if not self._gitCompareRepo( vcsRepo, remote_url ) :
                self._errorExit( "Git remote mismatch: '{0}' != '{1}'"
                    .format(vcsRepo, remote_url ) )

            self._callExternal( [ gitBin, 'fetch', '-q'  ] )
        else :
            self._callExternal( [ gitBin, 'clone', '-q', vcsRepo, wc_dir ] )
            
            try:
                self._callExternal( [ gitBin, 'rev-parse', 'HEAD' ], verbose=False )
            except subprocess.CalledProcessError:
                # exit on empty repository
                return
            
        remote_branch = self._callExternal( [
            gitBin, 'branch', '-q', '--all', '--list', 'origin/'+vcs_ref
        ], verbose=False ).strip()
            
        if self._callExternal( [ gitBin, 'branch', '-q', '--list', vcs_ref ], verbose=False ).strip():
            self._callExternal( [ gitBin, 'checkout', '-q', vcs_ref ] )

            if remote_branch:
                self._callExternal( [ gitBin, 'branch', '-q', '--set-upstream-to', 'origin/'+vcs_ref ] )
                self._callExternal( [ gitBin, 'rebase', 'origin/'+vcs_ref  ] )
        elif remote_branch:
            self._callExternal( [ gitBin, 'checkout', '-q', '--track', '-b', vcs_ref, 'origin/'+vcs_ref ] )
        else:
            self._callExternal( [ gitBin, 'checkout', '-q', vcs_ref ] )
    
    def vcsCommit( self, config, message, files ):
        env = config['env']
        gitBin = env['gitBin']
        self._checkGitConfig(env)

        if files:
            self._callExternal( [ gitBin, 'add' ] + files )
        else:
            self._callExternal( [ gitBin, 'add', '-A' ] )
            files = []

        self._callExternal( [ gitBin, 'commit', '-q', '-m', message ] + files )
    
    def vcsTag( self, config, tag, message ):
        gitBin = config['env']['gitBin']
        self._callExternal( [ gitBin, 'tag', '-a', '-m', message, tag ] )
    
    def vcsPush( self, config, refs, repo=None, check_empty=True ):
        refs = refs or []
        gitBin = config['env']['gitBin']
        repo = repo or 'origin'
        
        if check_empty:
            try:
                self._callExternal( [ gitBin, 'rev-parse', 'HEAD' ], verbose=False )
            except subprocess.CalledProcessError:
                # exit on empty repository
                return

        self._callExternal( [ gitBin, '-c', 'push.default=current', 'push', '-q', repo ] + refs )
        
    def vcsGetRevision( self, config ) :
        gitBin = config['env']['gitBin']
        return self._callExternal( [
            gitBin, 'rev-parse', 'HEAD'
        ], verbose=False ).strip()
    
    def vcsGetRefRevision( self, config, vcs_cache_dir, branch ) :
        res = self._callExternal( [
            config['env']['gitBin'],
            'ls-remote', '--refs',
            config['vcsRepo'],
            'refs/heads/{0}'.format(branch)
        ], verbose=False ).strip()
        
        if res:
            return res.split()[0]
        
        self._errorExit( "Uknown Git ref: '{0}'".format( branch ) )
    

    def vcsListTags( self, config, vcs_cache_dir, tag_hint ) :
        if tag_hint:
            tag_hint = ['refs/tags/{0}'.format(tag_hint)]
        else:
            tag_hint = []
        
        res = self._callExternal( [
            config['env']['gitBin'],
            'ls-remote','--tags', '--refs',
            config['vcsRepo']
        ] + tag_hint, verbose=False ).strip().split("\n")

        res = [ v and v.split()[1].replace('refs/tags/', '') or '' for v in res ]
        res = list(filter(None, res))
        return res
    
    def vcsListBranches( self, config, vcs_cache_dir, branch_hint ) :
        if branch_hint:
            branch_hint = ['refs/heads/{0}'.format(branch_hint)]
        else:
            branch_hint = []
        
        res = self._callExternal( [
            config['env']['gitBin'],
            'ls-remote','--heads', '--refs',
            config['vcsRepo']
        ] + branch_hint, verbose=False ).strip().split("\n")

        res = [ v and v.split()[1].replace('refs/heads/', '') or '' for v in res ]
        res = list(filter(None, res))
        return res

    def vcsExport( self, config, vcs_cache_dir, vcs_ref, dst_path ) :
        env = config['env']
        gitBin = env['gitBin']
        vcsRepo = config['vcsRepo']

        if vcs_cache_dir is None:
            cache_repo = vcsRepo
        else:
            if os.path.exists(vcs_cache_dir):
                remote_url = self.vcsGetRepo( config, vcs_cache_dir )
                
                if not self._gitCompareRepo(vcsRepo,  remote_url):
                    self._warn('removing git cache on remote URL mismatch: {0} != {1}'
                        .format(remote_url, vcsRepo))
                    self._rmTree(vcs_cache_dir)
            
            if not os.path.exists(vcs_cache_dir):
                self._callExternal( [
                    env['gitBin'],
                    'clone',
                    '--mirror',
                    '--depth=1',
                    '--no-single-branch',
                    vcsRepo,
                    vcs_cache_dir
                ] )
            else:
                self._callExternal( [
                    env['gitBin'],
                    '--git-dir={0}'.format(vcs_cache_dir),
                    'fetch'
                ] )
                
            cache_repo = 'file://' + vcs_cache_dir
            
        if os.path.exists(dst_path):
            self._rmTree(dst_path)

        os.mkdir(dst_path)
        
        self._callBash(env, '{0} archive --remote={1} --format=tar {2} | {3} x -C {4}'
                .format(config['env']['gitBin'], cache_repo, vcs_ref, env['tarBin'], dst_path))

    def vcsBranch( self, config, vcs_ref ):
        env = config['env']
        gitBin = env['gitBin']
        
        self._callExternal( [
            config['env']['gitBin'],
            'checkout', '-b', vcs_ref,
        ] )
        
        self.vcsPush(config, [vcs_ref])

    def vcsMerge( self, config, vcs_ref, cleanup ):
        curr_ref = self._getCurrentBranch( config )
        
        env = config['env']
        gitBin = env['gitBin']
        
        try:
            self._callExternal( [
                config['env']['gitBin'],
                'merge', '--no-ff', 'origin/'+vcs_ref,
            ] )
        except subprocess.CalledProcessError:
            if cleanup:
                self.vcsRevert(config)
                self._errorExit('Merge failed, aborted.')
            self._errorExit('Merge failed, left as-is.')
        
        self.vcsPush(config, [curr_ref])

    def vcsDelete( self, config, vcs_cache_dir, vcs_ref ):
        env = config['env']
        gitBin = env['gitBin']
        
        have_local = (os.path.exists('.git') and
                      self._gitCompareRepo(config['vcsRepo'],  self.vcsGetRepo(config)))
        
        if have_local:
            try:
                self._callExternal( [
                    config['env']['gitBin'],
                    'branch', '-D', vcs_ref,
                ] )
            except subprocess.CalledProcessError as e:
                self._warn(str(e))
                
            self.vcsPush(config, ['--force', '--delete', vcs_ref])
                
            self._callExternal( [
                config['env']['gitBin'],
                'fetch', '--all', '--prune',
            ] )
        else:
            # a dirty hack to avoid error message of not found
            # local git repo
            repo = tempfile.mkdtemp('fakegit', dir='.')
            
            self._callExternal( [
                config['env']['gitBin'],
                'init', repo,
            ], verbose=False )

            oldcwd = os.getcwd()
            os.chdir(repo)
            self.vcsPush(config, ['--force', '--delete', vcs_ref], config['vcsRepo'], False)
            os.chdir(oldcwd)

            self._rmTree(repo)

    def vcsRevert( self, config):
        if os.path.exists(os.path.join('.git', 'MERGE_HEAD')):
            self._callExternal( [
                config['env']['gitBin'],
                'merge', '--abort',
            ] )

        self._callExternal( [
            config['env']['gitBin'],
            'reset', '--hard',
        ] )

    def vcsIsMerged( self, config, vcs_ref ):
        res = self._callExternal( [
            config['env']['gitBin'],
            'branch', '-r', '--merged', 'HEAD', 'origin/{0}'.format(vcs_ref)
        ], verbose=False ).strip()
        return res != ''
    




from __future__ import print_function, absolute_import

import hashlib

from .subtool import SubTool
from .coloring import Coloring

__all__ = ['RmsTool']

class RmsTool( SubTool ):
    def autoDetect( self, config ) :
        return self._autoDetectRMS( config )
    
    def rmsPromoteMany( self, config, packages, rms_pool ):
        for p in packages:
            self.rmsPromote( config, p, rms_pool)
    
    def rmsPromote( self, config, package, rms_pool ):
        raise NotImplementedError( self._name )

    def rmsGetList( self, config, rms_pool, package_hint ):
        raise NotImplementedError( self._name )
    
    def rmsRetrieve( self, config, rms_pool, package ):
        raise NotImplementedError( self._name )
    
    def _autoDetectRMS( self, config ) :
        if config.get( 'rms', None ) == self._name :
            return True
        
        return False

    @classmethod
    def rmsCalcHash( cls, file_name, hash_type ) :
        hf = hashlib.new( hash_type )
        with open( file_name, 'rb' ) as f:
            for chunk in iter(lambda: f.read(4096), ''):
                if not chunk: break
                hf.update( chunk )
        return "{0}:{1}".format( hash_type, hf.hexdigest() )

    @classmethod
    def rmsVerifyHash( cls, config, file_name ) :
        if config.get('rmsHash', None) is not None:
            ( hash_type, hash_value ) = config['rmsHash'].split(':',1)
            file_hash = cls.rmsCalcHash( file_name, hash_type )
            
            if file_hash != config['rmsHash'] :
                raise RuntimeError(
                        "RMS hash mismatch {0} != {1}".format(
                            file_hash, hash_value ) )
        else :
            file_hash = cls.rmsCalcHash( file_name, 'sha256' )
            
            print( Coloring.label("File: ") + file_name )
            print( Coloring.label("Hash: ") + file_hash )
            yn = raw_input( "Is it correct? (Y/N) " )
            
            if yn not in ( 'Y', 'y' ):
                raise RuntimeError( 'User abort on RMS hash validation' )






import os, subprocess

from ..runenvtool import RunEnvTool
from .bashtoolmixin import BashToolMixIn

class rustupTool( BashToolMixIn, RunEnvTool ):
    """rustup is an installer for the systems programming language Rust.

Home: https://www.rustup.rs/
"""
    DIR_DEFAULT = os.path.join(os.environ['HOME'], '.multirust')
    CARGO_DIR_DEFAULT = os.path.join(os.environ['HOME'], '.cargo')
    INSTALLER_DEFAULT = 'https://sh.rustup.rs'

    def getDeps( self ) :
        return [ 'bash', 'curl' ]

    def _installTool( self, env ):
        self._callBash( env,
              '{0} {1} -sSf | {2} -s -- -y --no-modify-path'
               .format(env['curlBin'], env['rustupInstaller'], env['bashBin'])
        )
            
    def updateTool( self, env ):
        self._callExternal([
            env['rustupBin'], 'self', 'update'
        ])
    
    def uninstallTool( self, env ):
        for v in ['rustupDir', 'cargoDir']:
            dir = env[v]
            
            if os.path.exists(dir):
                self._rmTree(dir)
            
        self._have_tool = False
            
    def envNames( self ) :
        return ['rustupBin', 'rustupDir', 'rustupInstaller']
    
    def initEnv( self, env ) :
        dir = os.environ.setdefault('RUSTUP_HOME', self.DIR_DEFAULT)
        cargo_dir = os.environ.setdefault('CARGO_HOME', self.CARGO_DIR_DEFAULT)
        
        dir = env.setdefault('rustupDir', dir)
        cargo_dir = env.setdefault('cargoDir', cargo_dir)
        
        os.environ['RUSTUP_HOME'] = dir
        os.environ['CARGO_HOME'] = cargo_dir
        
        env.setdefault('rustupInstaller', self.INSTALLER_DEFAULT)

        bin_dir = os.path.join(cargo_dir, 'bin')
        self._addBinPath(bin_dir, True)
        
        if os.path.exists(os.path.join(bin_dir, 'rustup')):
            super(rustupTool, self).initEnv( env )




from __future__ import print_function

import os, re, sys, subprocess

from ..vcstool import VcsTool
from ..rmstool import RmsTool

try:
    import xml.dom.minidom as minidom
except ImportError:
    print('WARNING: missing xml.dom.minidom - SVN will not work', file=sys.stderr)

class svnTool( VcsTool, RmsTool ):
    """Apache Subversion: Enterprise-class centralized version control for the masses.
    
Home: https://subversion.apache.org/
"""    
    def _installTool( self, env ):
        self._requirePackages(['subversion'])
        self._requireEmerge(['dev-vcs/subversion'])
        self._requirePacman(['subversion'])
        self._requireHomebrew('subversion')
        
    def autoDetectFiles( self ):
        return '.svn'
    
    def autoDetect( self, config ) :
        return ( VcsTool.autoDetect( self, config ) or
                 RmsTool.autoDetect( self, config ) )
    
    def vcsGetRepo( self, config, wc_dir=None ):
        wc_dir = wc_dir or os.getcwd()
        
        svn_info = self._callExternal([
            config['env']['svnBin'], 'info', '--xml', wc_dir
        ], verbose=False)

        svn_info = minidom.parseString( svn_info )
        svn_info = svn_info.getElementsByTagName('url')
        url = svn_info[0].firstChild.nodeValue

        url = re.sub( '/(trunk|branches|tags).*$', '', url )
        return url
    
    def _detectSVNPath( self, config, vcs_ref ):
        env = config['env']
        svnBin = env['svnBin']
        vcsRepo = config['vcsRepo']        
        branch_path = '{0}/branches/{1}'.format( vcsRepo, vcs_ref )
        tag_path = '{0}/tags/{1}'.format( vcsRepo, vcs_ref )
        
        if vcs_ref == 'trunk':
            svn_repo_path = '{0}/trunk'.format( vcsRepo )
        elif self._callExternal( [ svnBin, 'info', branch_path ], suppress_fail=True ) :
            svn_repo_path = branch_path
        elif self._callExternal( [ svnBin, 'info', tag_path ], suppress_fail=True ) :
            svn_repo_path = tag_path
        else:
            self._errorExit( "VCS ref was not found: " + vcs_ref )
            
        return svn_repo_path
    
    def vcsCheckout( self, config, vcs_ref ):
        env = config['env']
        svnBin = env['svnBin']
        wc_dir = os.getcwd()
        vcs_ref = vcs_ref or 'trunk'
        
        svn_repo_path = self._detectSVNPath(config, vcs_ref)

        if os.path.isdir( '.svn' ):
            self._callExternal( [ svnBin, 'switch', svn_repo_path ] )
        else:
            self._callExternal(
                [ svnBin, 'checkout',
                 svn_repo_path,
                 wc_dir ] )
    
    def vcsCommit( self, config, message, files ):
        svnBin = config['env']['svnBin']
        if files:
            commit_files = files
        else:
            commit_files = []
            files = ['--depth=infinity', '.']

        self._callExternal( [ svnBin, 'add', '--force', ] + files )

        self._callExternal(
                [ svnBin, 'commit',
                 '-m', message ] + commit_files )
    
    def vcsTag( self, config, tag, message ):
        env = config['env']
        svnBin = env['svnBin']

        svn_info = self._callExternal([
            config['env']['svnBin'], 'info', '--xml'
        ], verbose=False)
        svn_info = minidom.parseString( svn_info )
        svn_info = svn_info.getElementsByTagName('url')
        svn_url = svn_info[0].firstChild.nodeValue

        self._callExternal( [
            svnBin, 'copy',
            '-m', message,
            '--parents',
            svn_url,
            '%s/tags/%s' % ( config['vcsRepo'], tag )
        ] )
    
    def vcsPush( self, config, refs ):
        pass
        
    def vcsGetRevision( self, config ) :
        svnBin = config['env']['svnBin']
        svn_info = self._callExternal( [ svnBin, 'info',  '--xml' ], verbose=False )
        
        svn_info = minidom.parseString( svn_info )
        svn_info = svn_info.getElementsByTagName('commit')

        if len(svn_info):
            return svn_info[0].getAttribute('revision')
            
        return 'local'

    def vcsGetRefRevision( self, config, vcs_cache_dir, branch ) :
        svnBin = config['env']['svnBin']
        
        svn_repo_path = self._detectSVNPath(config, branch)
            
        res = self._callExternal( [ svnBin, 'info', svn_repo_path, '--xml' ], verbose=False )
        
        res = minidom.parseString( res )
        return res.getElementsByTagName('commit')[0].getAttribute('revision')

    def vcsListTags( self, config, vcs_cache_dir, tag_hint ) :
        return self._vcsListCommon(config, vcs_cache_dir, 'tags')
    
    def vcsListBranches( self, config, vcs_cache_dir, branch_hint ) :
        return self._vcsListCommon(config, vcs_cache_dir, 'branches')
    
    def _vcsListCommon( self, config, vcs_cache_dir, sub_path ):
        svnBin = config['env']['svnBin']
        vcsRepo = config['vcsRepo']
        svn_repo_path = '{0}/{1}/'.format( vcsRepo, sub_path )
        
        res = self._callExternal( [
            svnBin, 'ls', svn_repo_path
        ], verbose=False ).strip().split("\n")
        
        res = [v and v.replace(svn_repo_path, '').replace('/', '') or '' for v in res ]
        res = list(filter(None, res))
        return res

    def vcsExport( self, config, vcs_cache_dir, vcs_ref, dst_path ) :
        svnBin = config['env']['svnBin']
        
        svn_repo_path = self._detectSVNPath(config, vcs_ref)
        
        if os.path.exists(dst_path):
            os.rmdir(dst_path)
        
        if vcs_cache_dir is None:
            self._callExternal( [ svnBin, 'export', svn_repo_path, dst_path ] )
            return
            
        if os.path.isdir( vcs_cache_dir ):
            cnd = 'switch'
        else:
            cnd = 'checkout'
            
        self._callExternal( [ svnBin, cnd, svn_repo_path, vcs_cache_dir ] )
        self._callExternal( [ svnBin, 'export', vcs_cache_dir, dst_path ] )
    
    def vcsBranch( self, config, vcs_ref ):
        svnBin = config['env']['svnBin']
        vcsRepo = config['vcsRepo']
        svn_dst_path = '{0}/branches/{1}'.format( vcsRepo, vcs_ref )
        
        svn_info = self._callExternal([
            config['env']['svnBin'], 'info', '--xml'
        ], verbose=False)
        svn_info = minidom.parseString( svn_info )
        svn_info = svn_info.getElementsByTagName('url')
        svn_src_path = svn_info[0].firstChild.nodeValue
        
        if self._callExternal([ config['env']['svnBin'], 'info', svn_dst_path ], suppress_fail=True):
            self._errorExit('VCS branch {0} already exists!'.format(vcs_ref))

        self._callExternal( [
                svnBin, 'copy', '--parents',
                '-m', 'CID branch ' + vcs_ref,
                svn_src_path, svn_dst_path,
        ] )
        self.vcsCheckout(config, vcs_ref)

    def vcsMerge( self, config, vcs_ref, cleanup ):
        svnBin = config['env']['svnBin']
        
        svn_repo_path = self._detectSVNPath(config, vcs_ref)

        self._callExternal( [ svnBin, 'update' ] )
        self._callExternal( [ svnBin, 'merge', '--accept', 'postpone', svn_repo_path ] )
        
        try:
            self._callExternal( [ svnBin, 'commit', '-m', 'CID merged ' + vcs_ref ] )
        except subprocess.CalledProcessError:
            if cleanup:
                self.vcsRevert(config)
                self._errorExit('Merge failed, aborted.')
            self._errorExit('Merge failed, left as-is.')

    def vcsDelete( self, config, vcs_cache_dir, vcs_ref ):
        svnBin = config['env']['svnBin']
        vcsRepo = config['vcsRepo']
        svn_repo_path = '{0}/branches/{1}'.format( vcsRepo, vcs_ref )
        self._callExternal( [ svnBin, 'remove', '-m', 'CID delete ' + vcs_ref, svn_repo_path ] )

    def vcsRevert( self, config):
        svnBin = config['env']['svnBin']
        self._callExternal( [ svnBin, 'revert', '-R', '.' ] )

    def vcsIsMerged( self, config, vcs_ref ):
        vcsRepo = config['vcsRepo']
        svn_repo_path = '{0}/branches/{1}'.format( vcsRepo, vcs_ref )
        
        res = self._callExternal( [
            config['env']['svnBin'],
            'mergeinfo', '--show-revs',
            'eligible',
            svn_repo_path
        ], verbose=False ).strip()

        return res == ''




from ..buildtool import BuildTool
from ..testtool import TestTool
from .sdkmantoolmixin import SdkmanToolMixIn

class mavenTool( SdkmanToolMixIn, BuildTool, TestTool ):
    """Apache Maven is a software project management and comprehension tool.

Home: https://maven.apache.org/

Expects clean, compile, package and test targets.

Build targets:
    prepare -> clean
    build -> compile
    package -> package
    check -> test
Override targets with .config.toolTune.

Requires Java >= 7.
"""
    _MIN_JAVA = '7'
    
    def autoDetectFiles( self ) :
        return 'pom.xml'
    
    def _binName( self ):
        return 'mvn'

    def onPrepare( self, config ):
        target = self._getTune(config, 'prepare', 'clean')
        self._callExternal( [ config['env']['mavenBin'], target ] )
    
    def onBuild( self, config ):
        target = self._getTune(config, 'build', 'compile')
        self._callExternal( [ config['env']['mavenBin'], target ] )
        
    def onPackage( self, config ):
        target = self._getTune(config, 'package', 'package')
        self._callExternal( [ config['env']['mavenBin'], target ] )
        self._addPackageFiles(config, 'target/*.jar')

    def onCheck( self, config ):
        target = self._getTune(config, 'check', 'test')
        self._callExternal( [ config['env']['mavenBin'], target ] )

            



import os

from ..buildtool import BuildTool
from ..testtool import TestTool

class cargoTool( BuildTool, TestTool ):
    """Cargo, Rust;s Package Manager.
    
Home: http://doc.crates.io/

Build targets:
    prepare -> clean
    build -> build
    check -> test
Override targets with .config.toolTune.

"""
    def autoDetectFiles( self ) :
        return 'Cargo.toml'
        
    def getDeps( self ) :
        return [ 'rust' ]
        
    def uninstallTool( self, env ):
        pass
    
    def onPrepare( self, config ):
        self._callExternal( [
            config['env']['cargoBin'], 'clean',
        ])
    
    def onBuild( self, config ):
        args = []
        
        if not config.get('debugBuild', False):
            args.append('--release')
        
        
        self._callExternal( [ config['env']['cargoBin'], 'build' ] + args)
    
    def onPackage( self, config ):
        self._callExternal( [ config['env']['cargoBin'], 'package', '--allow-dirty' ])
        self._addPackageFiles(config, 'target/package/*.crate')

    def onCheck( self, config ):
        self._callExternal( [ config['env']['cargoBin'], 'test' ])
        
    def onRunDev( self, config ):
        self._callExternal( [ config['env']['cargoBin'], 'run' ])


        





import os, subprocess

from ..runenvtool import RunEnvTool
from .bashtoolmixin import BashToolMixIn

class sdkmanTool( BashToolMixIn, RunEnvTool ):
    """SDK Man for Java.

Home: http://sdkman.io/
"""
    SDKMAN_DIR_DEFAULT = os.path.join(os.environ['HOME'], '.sdkman')

    def getDeps( self ) :
        return [ 'bash', 'curl', 'unzip', 'zip' ]
            
    def envNames( self ) :
        return ['sdkmanDir', 'sdkmanGet']
    
    def _installTool( self, env ):
        dir = env['sdkmanDir']
        get = env.get('sdkmanGet', 'https://get.sdkman.io')

        self._callBash( env,
              'export SDKMAN_DIR="{2}"; {0} -s "{1}" | {3}'
               .format(env['curlBin'], get, dir, env['bashBin']) )
            
    def updateTool( self, env ):
        self._callBash( env,
            'source {0} && sdk selfupgrade'.format(env['sdkmanInit'])
        )
    
    def uninstallTool( self, env ):
        dir = env['sdkmanDir']

        if os.path.exists(dir):
            self._rmTree(dir)

        self._have_tool = False
    
    def initEnv( self, env ) :
        dir = env.setdefault('sdkmanDir', self.SDKMAN_DIR_DEFAULT)
        env_init = os.path.join(dir, 'bin', 'sdkman-init.sh')
        env['sdkmanInit'] = env_init
        self._have_tool = os.path.exists(env_init)
        
        
    def onExec( self, env, args ):
        self._callBashInteractive(env,
                '. {0} && sdk {1}'
                .format(env['sdkmanInit'], subprocess.list2cmdline(args))
        )




class JavaToolMixIn( object ):
    _LATEST_JAVA = '8'
    _MIN_JAVA = None
    
    def envDeps( self, env ):
        if not self._MIN_JAVA:
            return
        
        for ek in ['javaVer', 'jdkVer']:
            if env.get(ek, None):
                if int(env[ek]) < int(self._MIN_JAVA):
                    self._errorExit('{0} requires minimal {1}={2}'
                        .format(self._name, ek, self._MIN_JAVA))
            else:
                env[ek] = str(self._LATEST_JAVA)
            




import os, glob

from ..runenvtool import RunEnvTool
from .javatoolmixin import JavaToolMixIn

class jdkTool( RunEnvTool, JavaToolMixIn ):
    """Java Development Kit.
    
Home: http://openjdk.java.net/

Due to issues with Oracle's licensing, cid
supports only automatic installation of OpenJDK.

jdkVer supports:
- only one digit like 6, 7, 8, 9 (Zulu JDK (http://zulu.org)
- use jdkDir for custom JDK
- jdkVer is equal to javaVer by default
"""
    def getDeps( self ) :
        return ['java']
    
    def envNames( self ) :
        return ['jdkDir', 'jdkVer']
    
    def getVersionParts( self ):
        return 1
    
    def _installTool( self, env ):
        if env.get('jdkDir', None):
            return
        
        ver = env['jdkVer']
        
        # Zulu is installed in javaTool
        # leaving it here for possible future use
        #self._requireDeb(['openjdk-{0}-jdk'.format(ver)])
        #self._requireYum(['java-1.{0}.0-openjdk-devel'.format(ver)])
        #self._requireZypper(['java-1_{0}_0-openjdk-devel'.format(ver)])
        self._requirePacman(['jdk{0}-openjdk'.format(ver)])
        self._requireEmerge(['=dev-java/oracle-jdk-bin-1.{0}*'.format(ver)])
    
    def uninstallTool( self, env ):
        pass

    def initEnv( self, env ) :
        if env.get('jdkDir', None):
            java_home = env['jdkDir']
            bin_dir = os.path.join(java_home, 'bin')
            
            os.environ['JAVA_HOME'] = java_home
            os.environ['JDK_HOME'] = java_home
            
            self._addBinPath(bin_dir, True)
            
            super(jdkTool, self).initEnv( env, 'javac' )
            return
        
        env.setdefault('jdkVer', env['javaVer'])
        ver = env['jdkVer']
        
        candidates = [
            # Zulu
            '/usr/lib/jvm/zulu-{0}*/bin/javac'.format(ver),
            # Debian / Ubuntu
            #"/usr/lib/jvm/java-{0}-openjdk*/bin/javac".format(ver),
            # RedHat
            #"/usr/lib/jvm/java-1.{0}.0/bin/javac".format(ver),
            # OpenSuse
            #"/usr/lib*/jvm/java-1.{0}.0/bin/javac".format(ver),
            # Default oracle
            #"/opt/jdk/jdk1.{0}*/bin/javac".format(ver),
        ]
        
        if self._isGentoo() or self._isArchLinux():
            candidates += [
                "/usr/lib/jvm/java-{0}-openjdk*/bin/javac".format(ver),
            ]
        
        for c in candidates:
            bin_name = glob.glob(c)
            
            if bin_name:
                bin_name = bin_name[0]
                bin_dir = os.path.dirname(bin_name)
                java_home = os.path.dirname(bin_dir)
                
                env['jdkDir'] = java_home
                os.environ['JAVA_HOME'] = java_home
                os.environ['JDK_HOME'] = java_home
                
                self._addBinPath(bin_dir, True)
                self._have_tool = True
                break




from __future__ import print_function, absolute_import

class VersionMixIn( object ):
    """
CID tools are required to ignore bug-fix/patch release
version parts and always install the latest from release line.
"""    
    def getVersionParts( self ):
        """Override, if tool has different number of meaningful parts"""
        return 2
    
    def sanitizeVersion( self, env ):
        """Should be called implicitly by standard CID functionality"""
        ver_var = self._name + 'Ver'
        ver = env.get(ver_var, None)
        
        if ver:
            end = self.getVersionParts()
            new_ver = ver.split('.')[:end]
            new_ver = '.'.join(new_ver)
            
            if env[ver_var] != new_ver:
                self._warn('Too precise version "{0}" for "{1}" - trimmed to "{2}"'
                    .format(ver, self._name, new_ver))
                env[ver_var] = new_ver
            




from ..buildtool import BuildTool

class makeTool( BuildTool ):
    """GNU Make.
    
Home: https://www.gnu.org/software/make/

Expects presence of "clean" target.
Build uses the default target.
"""
    def autoDetectFiles( self ) :
        return [
            'GNUmakefile',
            'makefile',
            'Makefile',
        ]
    
    def _installTool( self, env ):
            self._requirePackages(['make'])
            self._requireEmerge(['sys-devel/make'])
            self._requirePacman(['make'])

    def onPrepare( self, config ):
        self._callExternal( [ config['env']['makeBin'], 'clean' ] )

    def onBuild( self, config ):
        target = self._getTune(config, 'build')
        
        if target:
            args = [target]
        else:
            args = []

        self._callExternal( [ config['env']['makeBin'] ] + args )







import os, re

from ..buildtool import BuildTool
from .sdkmantoolmixin import SdkmanToolMixIn

class gradleTool( SdkmanToolMixIn, BuildTool ):
    """Gradle Build Tool.

Home: https://gradle.org/

Build targets:
    prepare -> clean
    build -> <default> without explicit target
    package -> dists
    packageGlob -> '*.jar'
Override targets with .config.toolTune.

Requires Java >= 7.
"""
    _MIN_JAVA = '7'
    
    def autoDetectFiles( self ) :
        return 'build.gradle'
    
    def envDeps( self, env ):
        super(gradleTool, self).envDeps( env )
        
        gradlew_prop = 'gradle/wrapper/gradle-wrapper.properties'
        
        if os.path.exists(gradlew_prop):
            with open(gradlew_prop, 'r') as f:
                props = f.read()
                env['gradleVer'] = re.search('gradle-([0-9.]+)-bin.zip', props).group(1)

    def onPrepare( self, config ):
        target = self._getTune(config, 'prepare', 'clean')
        self._callExternal( [ config['env']['gradleBin'], '-q', '--no-daemon', target ] )
    
    def onBuild( self, config ):
        target = self._getTune(config, 'build')
        
        if target:
            args = [target]
        else:
            args = []
            
        self._callExternal( [ config['env']['gradleBin'], '-q', '--no-daemon' ] + args )

    def onPackage( self, config ):
        target = self._getTune(config, 'package', 'dists')
        self._callExternal( [ config['env']['gradleBin'], '-q', '--no-daemon', target ] )
        
        packageGlob = self._getTune(config, 'packageGlob', 'build/libs/*.jar')
        self._addPackageFiles(config, packageGlob)
            



from __future__ import print_function, absolute_import

import os, json, stat, shutil, sys, grp
from collections import OrderedDict
from ..coloring import Coloring

class UtilMixIn( object ):
    _FUTOIN_JSON = 'futoin.json'
    
    def _updateEnvFromOutput( self, env_to_set ):
        env_to_set = env_to_set.split( "\n" )

        for e in env_to_set:
            if not e: continue
            n, v = e.split('=', 1)
            os.environ[n] = v
            
    def autoDetectFiles( self ):
        return None

    def _autoDetectByCfg( self, config, file_name ) :
        if self._name in config.get( 'toolOrder', [] ) :
            return True
        
        if type( file_name ) is type( '' ):
            file_name = [ file_name ]
            
        root_list = config['projectRootSet']

        for f in file_name :
            if f in root_list :
                return True
        
        return False
    
    def _loadJSONConfig( self, file_name ):
        if os.path.exists(file_name):
            with open(file_name, 'r') as content_file:
                content = content_file.read()
                object_pairs_hook = lambda pairs: OrderedDict( pairs )
                return json.loads( content, object_pairs_hook=object_pairs_hook )
        else:
            return None
            


    def _updateJSONConfig( self, file_name, updater, indent=2, separators=(',', ': ') ) :
        content = self._loadJSONConfig( file_name )
        
        if content is None:
            return []
            
        updater( content )
        
        self._writeJSONConfig( file_name, content, indent, separators )
            
        return [ file_name ]
    
    def _writeJSONConfig( self, file_name, content, indent=2, separators=(',', ': ')):
        with open(file_name, 'w') as content_file:
            content =  json.dumps( content, indent=indent, separators=separators )
            content_file.write( content )
            content_file.write( "\n" )
    
    def _updateTextFile( self, file_name, updater ) :
        with open(file_name, 'r') as content_file:
            content = content_file.read()
            
        content = updater( content )
        
        with open(file_name, 'w') as content_file:
            content_file.write( content )
            
        return [ file_name ]
    
    def _isExternalToolsSetup( self, env ):
        return env['externalSetup']['installTools']
    
    def _rmTree( self, dir ):
        print( Coloring.infoLabel('Removing: ') + Coloring.info(dir),
                file=sys.stderr )
        
        os.chmod( dir, stat.S_IRWXU )
        for ( path, dirs, files ) in os.walk( dir ) :
            for id in dirs + files :
                os.chmod( os.path.join( path, id ), stat.S_IRWXU )
        shutil.rmtree(dir)
        
    def _getTune( self, config, key, default=None ):
        return config.get('toolTune', {}).get(self._name, {}).get(key, default)
    
    def _info( self, msg ):
        print(Coloring.info('INFO: '+msg), file=sys.stderr)
    
    def _warn( self, msg ):
        print(Coloring.warn('WARNING: '+msg), file=sys.stderr)
    
    def _errorExit( self, msg ):
        raise RuntimeError(msg)
        
    def _haveGroup( self, grpname ):
        gid = grp.getgrnam(grpname)[2]
        return gid in os.getgroups()
        
        




from ..buildtool import BuildTool
from .sdkmantoolmixin import SdkmanToolMixIn

class antTool( SdkmanToolMixIn, BuildTool ):
    """Ant build tool for Java applications.
    
Home: http://ant.apache.org/

The tool assumes the following targets: clean, compile, jar, run

Ant is setup through SDKMan!

Note: If detected Java version is less than 8 then Ant 1.9.8 is used.

Build targets:
    prepare -> clean
    build -> compile
    package -> jar
Override targets with .config.toolTune.

"""
    def autoDetectFiles( self ) :
        return [ 'build.xml' ]
    
    def initEnv( self, env ) :
        if self._javaVersion( env ) < 8:
            env['antVer'] = '1.9.8'

        super(antTool, self).initEnv(env)

    def onPrepare( self, config ):
        target = self._getTune(config, 'prepare', 'clean')
        self._callExternal( [ config['env']['antBin'], target ] )
    
    def onBuild( self, config ):
        target = self._getTune(config, 'build', 'compile')
        self._callExternal( [ config['env']['antBin'], target ] )
        
    def onPackage( self, config ):
        target = self._getTune(config, 'package', 'jar')
        self._callExternal( [ config['env']['antBin'], target ] )
        
        self._addPackageFiles(config, 'build/jar/*.jar')
        
    def onRunDev( self, config ):
        self._callExternal( [ config['env']['antBin'], 'run' ] )

            



import os

from ..runenvtool import RunEnvTool

class gccTool( RunEnvTool ):
    """GNU Compiler Collection.
    
Home: https://gcc.gnu.org/

If gccBin is set it must point to bin folder.
"""    
    def envNames(self):
        return ['gccBin', 'gccPrefix', 'gccPostfix']
        
    def _installTool( self, env ):
        self._requireDeb(['gcc'])
        self._requireRpm(['gcc'])
        self._requireEmerge(['sys-devel/gcc'])
        self._requirePacman(['gcc'])
        
    def initEnv(self, env):
        gcc_bin = env.get('gccBin', None)
        
        pref = env.setdefault('gccPrefix', '')
        postf = env.setdefault('gccPostfix', '')
        
        
        if gcc_bin:
            if os.path.exists(gcc_bin):
                self._have_tool = True
        else:
            gcc = pref + 'gcc' + postf
            gcc = self._which(gcc)
            
            if gcc:
                env['gccBin'] = gcc
                self._have_tool = True
            



import os

from ..runenvtool import RunEnvTool

class sshTool( RunEnvTool ):
    """Secure Shell client.
    
Home: https://www.openssh.com/
"""    
    def _installTool( self, env ):
        self._requireDeb(['openssh-client'])
        self._requireRpm(['openssh'])
        self._requireEmerge(['virtual/ssh'])
        self._requirePacman(['openssh'])
    
    



import os

from ..buildtool import BuildTool
from .gemtoolmixin import GemToolMixIn


class puppetTool( GemToolMixIn, BuildTool ):
    """Puppet system automation.
    
Home: https://puppet.com/

Primary purpose is to support Puppet module development.
"""    
    METADATA_FILE = 'metadata.json'

    def envNames( self ) :
        return ['puppetVer', 'puppetBin' ]
    
    def initEnv( self, env ):
        super(puppetTool, self).initEnv( env )
        puppet_ver = env.setdefault('puppetVer', None)
        
        if self._have_tool and puppet_ver:
            try:
                found_ver = self._callExternal( [ env['puppetBin'], '--version' ], verbose = False )
                self._have_tool = found_ver.find(puppet_ver) >= 0
            except:
                self._have_tool = False
                del env['puppetBin']

    def autoDetectFiles( self ) :
        return self.METADATA_FILE

    def loadConfig( self, config ) :
        content = self._loadJSONConfig( self.METADATA_FILE )
        if content is None: return
        
        for f in ( 'name', 'version' ):
            if f in content and f not in config:
                config[f] = content[f]

        config['package'] = 'pkg'

    def updateProjectConfig( self, config, updates ) :
        def updater( json ):
            for f in ( 'name', 'version' ):
                if f in updates :
                        json[f] = updates[f]

        return self._updateJSONConfig( self.METADATA_FILE, updater )
    
    def onPrepare(self, config):
        if os.path.exists('pkg'):
            self._rmTree('pkg')

    def onBuild( self, config ):
        puppetBin = config['env']['puppetBin']
        self._callExternal( [puppetBin, 'module', 'build'] )

    def onPackage( self, config ):
        package_file = 'pkg/{0}-{1}.tar.gz'.format(
                config['name'],
                config['version']
        )

        if not os.path.exists( package_file ) :
            self._errorExit( 'Puppet Module built package is missing: ' + package_file )

        self._addPackageFiles(config, package_file)




from ..runtimetool import RuntimeTool

class exeTool( RuntimeTool ):
    """Dummy tool to execute files directly"""
    
    def initEnv(self, env):
        self._have_tool = True
    
    def onRun( self, config, file, args, tune ):
        env = config['env']
        self._callInteractive([file] + args)




import os

from ..buildtool import BuildTool
from .bashtoolmixin import BashToolMixIn

class composerTool( BashToolMixIn, BuildTool ):
    """Dependency Manager for PHP.
    
Home: https://getcomposer.org/

Composer is installed in composerDir as single Phar with "composer" name without
extension.
composerDir is equal to user's ~/bin/ folder by default.
"""
    COMPOSER_JSON = 'composer.json'
    
    def _installTool( self, env ):
        composer_dir = env['composerDir']
        php_bin = env['phpBin']
        curl_bin = env['curlBin']
        composer_get = env.get('composerGet', 'https://getcomposer.org/installer')

        self._callBash(
            env,
            'mkdir -p {2} &&  {3} -s {0} | {1} -- --install-dir={2} --filename=composer'
                .format(composer_get, php_bin, composer_dir, curl_bin)
        )

    def updateTool( self, env ):
        self._callExternal([env['composerBin'], 'self-update'])

    def uninstallTool( self, env ):
        os.remove(env['composerBin'])
        self._have_tool = False
        
    def envNames( self ) :
        return [ 'composerDir', 'composerBin', 'composerGet' ]

    def initEnv( self, env ) :
        bin_dir = env['binDir']
        composer_dir = env.setdefault('composerDir', bin_dir)
        composer_bin = env.setdefault('composerBin', os.path.join(composer_dir, 'composer'))

        self._addBinPath( composer_dir )

        self._have_tool = os.path.exists( composer_bin )

    def autoDetectFiles( self ) :
        return self.COMPOSER_JSON
    
    def getDeps( self ) :
        return [ 'php', 'curl', 'bash' ]

    def loadConfig( self, config ) :
        content = self._loadJSONConfig( self.COMPOSER_JSON )
        if content is None: return

        for f in ( 'name', 'version' ):
            if f in content and f not in config:
                config[f] = content[f]

    def updateProjectConfig( self, config, updates ) :
        def updater( json ):
            for f in ( 'name', 'version' ):
                if f in updates :
                        json[f] = updates[f]

        return self._updateJSONConfig( self.COMPOSER_JSON, updater, indent=4 )
    
    def onPrepare( self, config ):
        composerBin = config['env']['composerBin']
        self._callExternal( [ composerBin, 'install' ] )
        
    def onPackage( self, config ):
        composerBin = config['env']['composerBin']
        self._callExternal( [ composerBin, 'install', '--no-dev', '--no-scripts' ] )





import glob, subprocess

from ..runtimetool import RuntimeTool
from .bashtoolmixin import BashToolMixIn

class rubyTool( BashToolMixIn, RuntimeTool ):
    """Ruby is a dynamic, open source programming language.
    
Home: https://www.ruby-lang.org/en/

By default the latest available Ruby binary is used for the following OSes:
* Debian & Ubuntu - uses Brightbox builds 1.9, 2.0, 2.1, 2.2, 2.3.
* CentOS, RHEL & Oracle Linux - uses SCL 1.9, 2.0, 2.2, 2.3

You can forbid source builds by setting rubyBinOnly to non-empty string.

Otherwise, System Ruby is used by default.

If rubyVer is set then RVM is used to setup custom rubies.
That may lead to long time and resource consumption due to compilation,
if binary versions are not found for specific system.
"""
    def getDeps( self ) :
        return [ 'rvm' ]

    def _installTool( self, env ):
        ruby_ver = env['rubyVer']
        
        if ruby_ver == self.SYSTEM_VER:
            self._systemDeps()
            return
        
        if env['rubyBinOnly']:
            self._installBinaries( env )
            return

        self._buildDeps()
        self._callExternal([
            env['rvmBin'], 'install', ruby_ver, '--autolibs=read-only'
        ])
        self._callExternal([
            env['rvmBin'], 'cleanup', 'all'
        ])
        
    def _installBinaries( self, env ):
        ver = env['rubyVer']
        pkgver = ver
        
        if self._isDebian() or  self._isUbuntu():
            repo = env['rubyBrightboxRepo']
            
            self._addAptRepo(
                'brightbox-ruby',
                "deb {0} $codename$ main".format(repo),
                self._GPG_BIRGHTBOX_REPO,
                codename_map = {
                    # Ubuntu
                    'zesty': 'yakkety',
                    # Debian
                    'jessie': 'trusty',
                    'stretch': 'xenial',
                    'testing': 'yakkety',
                    'sid': 'yakkety',
                },
                repo_base = '{0}/dists'.format(repo)
            )
                
            if ver == '1.9': pkgver = '1.9.[0-9]'
            self._requireDeb([
                'ruby{0}'.format(pkgver),
                'ruby{0}-dev'.format(pkgver),
            ])
            ruby_bin = glob.glob('/usr/bin/ruby{0}*'.format(ver))[0]
            
        elif self._isSCLSupported():
            sclname = self._rubySCLName(ver)
            
            self._requireSCL()
            
            self._requireYum(sclname)
            
            # required for LD_LIBRARY_PATH
            env_to_set = self._callExternal(['scl', 'enable', sclname, 'env'], verbose=False)
            self._updateEnvFromOutput(env_to_set)
            
            ruby_bin = self._callExternal(['scl', 'enable', sclname, 'which ruby'], verbose=False).strip()
            
        else:
            self._systemDeps()
            return

        self._callExternal([
            env['rvmBin'], 'remove', 'ext-system-{0}'.format(ver)
        ], suppress_fail=True)
        
        self._callExternal([
            env['rvmBin'], 'mount', ruby_bin, '-n', 'system-{0}'.format(ver)
        ])
        
    def _rubySCLName( self, ver ):
        pkgver = ver.replace('.', '')
        
        if pkgver == '19':
            sclname = 'ruby193'
        elif pkgver == '20':
            sclname = 'ruby200'
        else:
            sclname = 'rh-ruby'+pkgver

        return sclname

    def updateTool( self, env ):
        if env['rubyVer'] != self.SYSTEM_VER and not env['rubyBinOnly']:
            self._installTool( env )
        
    def uninstallTool( self, env ):
        ruby_ver = env['rubyVer']
        
        if ruby_ver != self.SYSTEM_VER and not env['rubyBinOnly']:
            self._callExternal([
                env['rvmBin'], 'uninstall', env['rubyVer']
            ])
            self._have_tool = False

    def envNames( self ) :
        return ['rubyVer', 'rubyBin', 'rubyBinOnly' ]

    def initEnv( self, env ) :
        #---
        if self._isDebian() or self._isUbuntu():
            env.setdefault('rubyBrightboxRepo', 'http://ppa.launchpad.net/brightbox/ruby-ng/ubuntu')
            ruby_binaries = ['1.9', '2.0', '2.1', '2.2', '2.3']
        elif self._isSCLSupported():
            ruby_binaries = ['1.9', '2.0', '2.2', '2.3']
        else :
            ruby_binaries = None
            
        #---
        if ruby_binaries:
            ruby_ver = env.setdefault('rubyVer', ruby_binaries[-1])
            rubyBinOnly = ruby_ver in ruby_binaries
            
            rvm_ruby_ver = ruby_ver
            
            if rubyBinOnly:
                rvm_ruby_ver = 'ext-system-{0}'.format(ruby_ver)
                
                # required for LD_LIBRARY_PATH
                if self._isSCLSupported():
                    sclname = self._rubySCLName(ruby_ver)
                    
                    try:
                        env_to_set = self._callExternal(['scl', 'enable', sclname, 'env'], verbose=False)
                        self._updateEnvFromOutput(env_to_set)
                    except subprocess.CalledProcessError:
                        pass
        else:
            rubyBinOnly = False
            ruby_ver = env.setdefault('rubyVer', self.SYSTEM_VER)
            rvm_ruby_ver = ruby_ver
            
        rubyBinOnly = env.setdefault('rubyBinOnly', rubyBinOnly)
        
            
        
        #---
        rvm_dir = env['rvmDir']

        try:
            env_to_set = self._callBash( env,
                'source {0} && \
                rvm use {1} >/dev/null && \
                env | grep "rvm"'.format(env['rvmInit'], rvm_ruby_ver),
                verbose = False
            )
        except:
            return

        if env_to_set :
            self._updateEnvFromOutput(env_to_set)
            super(rubyTool, self).initEnv( env )
            
    def _buildDeps( self ):
        # APT
        #---
        self._requireDeb([
            'build-essential',
            'gawk',
            'libssl-dev',
            'make',
            'libc6-dev',
            'zlib1g-dev',
            'libyaml-dev',
            'libsqlite3-dev',
            'sqlite3',
            'autoconf',
            'libgmp-dev',
            'libgdbm-dev',
            'libncurses5-dev',
            'automake',
            'libtool',
            'bison',
            'pkg-config',
            'libffi-dev',
            'libgmp-dev',
            'libreadline-dev',
        ])
        

        # Extra repo before the rest
        #---
        self._requireYumEPEL()

        self._requireRpm([
            'binutils',
            'patch',
            'libyaml-devel',
            'autoconf',
            'gcc',
            'gcc-c++',
            'glibc-devel',
            'readline-devel',
            'zlib-devel',
            'libffi-devel',
            'openssl-devel',
            'automake',
            'libtool',
            'bison',
            'sqlite-devel',
            'make',
            'm4',
            'gdbm-devel',
            'libopenssl-devel',
            'sqlite3-devel',
        ])
        
        self._requireEmergeDepsOnly(['dev-lang/ruby'])
        self._requirePacman(['ruby'])
        
    def _systemDeps( self ):
        self._requirePackages(['ruby'])
        self._requireEmerge(['dev-lang/ruby'])
        self._requirePacman(['ruby'])
        
    _GPG_BIRGHTBOX_REPO = '''
-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: SKS 1.1.6
Comment: Hostname: keyserver.ubuntu.com

mI0ETKTCMQEEAMX3ttL4YFO5AQ7Z6L5gaGw57CJBQl6jCv6lka0p8DaGNkeX0Rs9DhINa8qR
hxJCPK6ijeoNss69G/ni+sMSRViJBFWXzitEE1ew5YM2sw7wLE3guToDu60kaDwIn5mR3GTx
cgqDrQeCuGZJgz3e2lgmGYw2rAhMe78rRgkR5GFvABEBAAG0G0xhdW5jaHBhZCBQUEEgZm9y
IEJyaWdodGJveIi4BBMBAgAiBQJMpMIxAhsDBgsJCAcDAgYVCAIJCgsEFgIDAQIeAQIXgAAK
CRD12l8Jwxc6pl2BA/4p5DFEpGVvkgLj7/YLYCtYmZDw8i/drGbkWfIQiOgPWIf8QgpJXVME
1tkH8N1ssjbJlUKl/HubNBKZ6HDyQsQASFug+eI6KhSFMScDBf/oMX3zVCTTvUkgJtOWYc5d
77zJacEUGoSEx63QUJVvp/LAnqkZbt17JJL6HOou/CNicw==
=G8vE
-----END PGP PUBLIC KEY BLOCK-----
'''




class GemToolMixIn( object ):
    def getDeps( self ) :
        return [ 'gem', 'ruby' ]
    
    def _gemName( self ):
        return self._name
        
    def _installTool( self, env ):
        puppet_ver = env.get(self._name + 'Ver', None)
        version_arg = []
        
        if puppet_ver:
            version_arg = ['--version', puppet_ver]

        self._callExternal( [ env['gemBin'], 'install', self._gemName() ] + env['gemInstallArgs'].split(' ') + version_arg )
        
    def updateTool( self, env ):
        if env.get(self._name + 'Ver', None) :
            self._installTool( self, env )
        else :
            self._callExternal( [ env['gemBin'], 'update', self._gemName() ] + env['gemInstallArgs'].split(' ') )
        
    def uninstallTool( self, env ):
        self._callExternal( [ env['gemBin'], 'uninstall', self._gemName() ] )
        self._have_tool = False




from ..runtimetool import RuntimeTool
from .sdkmantoolmixin import SdkmanToolMixIn

class scalaTool( SdkmanToolMixIn, RuntimeTool ):
    """The Scala Programming Language.
    
Home: https://www.scala-lang.org/

Installed via SDKMan!

Requires Java >= 8.
"""
    _MIN_JAVA = '8'




import os

from ..runenvtool import RunEnvTool

class binutilsTool( RunEnvTool ):
    """GNU Binutils.
    
Home: https://www.gnu.org/software/binutils/

If binutilsDir is set it must point to bin folder.
"""    
    def envNames(self):
        return ['binutilsDir', 'binutilsPrefix', 'binutilsPostfix']
        
    def _installTool( self, env ):
        self._requireDeb(['binutils'])
        self._requireRpm(['binutils'])
        self._requireEmerge(['sys-devel/binutils'])
        self._requirePacman(['binutils'])
        
    def initEnv(self, env):
        pref = env.setdefault('binutilsPrefix', '')
        postf = env.setdefault('binutilsPostfix', '')
        
        bu_dir = env.get('binutilsDir', None)
        ld = pref + 'ld' + postf
        
        if bu_dir:
            ld = os.path.join(bu_dir, ld)
            self._have_tool = os.path.exists(bu_dir)
        else:
            ld = self._which(ld)
            
            if ld:
                env['binutilsDir'] = os.path.dirname(ld)
                self._have_tool = True
            



import os
import subprocess
import tempfile
import glob

from ..vcstool import VcsTool
from .bashtoolmixin import BashToolMixIn

class hgTool( BashToolMixIn, VcsTool ):
    """Mercurial SCM.
    
Home: https://www.mercurial-scm.org/
"""
    def getDeps( self ) :
        return [ 'bash' ]
    
    def _installTool( self, env ):
        self._requirePackages(['mercurial'])
        self._requireEmerge(['dev-vcs/mercurial'])
        self._requirePacman(['mercurial'])
        self._requireHomebrew('mercurial')
    
    def autoDetectFiles( self ) :
        return '.hg'

    def _getCurrentBranch( self, config ):
        return self._callExternal( [
            config['env']['hgBin'], 'branch'
        ] ).strip()

    def vcsGetRepo( self, config, wc_dir=None ):
        return self._callExternal( [
            config['env']['hgBin'], '--repository', wc_dir or os.getcwd(), 'paths', 'default'
        ], verbose=False ).strip()
    
    def _hgCheckoutTool( self, config ):
        hgBin = config['env']['hgBin']
        help_res = self._callExternal( [ hgBin, 'checkout', '--help' ], verbose=False )
        tool_args = []
        
        if help_res.find('--tool') > 0:
            tool_args += ['--tool', 'internal:merge']
            
        return tool_args
        
    def vcsCheckout( self, config, vcs_ref ):
        hgBin = config['env']['hgBin']
        wc_dir = os.getcwd()

        if os.path.isdir( '.hg' ):
            if 'vcsRepo' in config:
                remote_info = self.vcsGetRepo( config, '.' )
                if remote_info != config['vcsRepo'] :
                    self._errorExit( "Hg remote mismatch: " + remote_info )

            self._callExternal( [ hgBin, 'pull' ] )
        else :
            self._callExternal( [ hgBin, 'clone', config['vcsRepo'], wc_dir ] )
            
        if not vcs_ref:
            # skip default branch
            return
            
        for v in self._callExternal( [ hgBin, 'branches' ], verbose=False ).strip().split("\n"):
            if v and v.split()[0] == vcs_ref:
                break
        else:
            for v in self._callExternal( [ hgBin, 'tags' ], verbose=False ).strip().split("\n"):
                if v and v.split()[0] == vcs_ref:
                    break
            else:
                self._errorExit('Unknown VCS ref {0}. Hint: closed branches are ignored!'.format(vcs_ref))
            
        self._callExternal( [ hgBin, 'checkout', '--check', vcs_ref ] + self._hgCheckoutTool(config) )
    
    def vcsCommit( self, config, message, files ):
        hgBin = config['env']['hgBin']
        files = files or ['-A']
        self._callExternal( [ hgBin, 'commit', '-A', '-m', message ] + files )
    
    def vcsTag( self, config, tag, message ):
        hgBin = config['env']['hgBin']
        self._callExternal( [ hgBin, 'tag', '-m', message, tag ] )
    
    def vcsPush( self, config, refs ):
        refs = refs or []
        env = config['env']
        hgBin = env['hgBin']
        opts = []
        
        tags = self.vcsListTags(config, '.', '')
        
        for r in refs :
            if r in tags : continue
            if r == '--new-branch':
                opts.append( r )
                continue
                
            opts.append( '-b' )
            opts.append( r )
        self._callExternal( [ hgBin, 'push' ] + opts )
        
    def vcsGetRevision( self, config ) :
        hgBin = config['env']['hgBin']
        return self._callExternal( [ hgBin, 'identify', '--id' ], verbose=False ).strip()
    
    def _hgCache( self, config, vcs_cache_dir ):
        hgBin = config['env']['hgBin']
        vcsrepo = config['vcsRepo']
        
        if vcs_cache_dir is None:
            # Hg does not allow remote repository listing
            if os.path.exists('.hg') and self.vcsGetRepo( config, vcs_cache_dir ) == vcsrepo:
                vcs_cache_dir = '.'
            else:
                vcs_cache_dir = vcsrepo.replace('/', '_').replace(':', '_')
            vcs_cache_dir = os.path.realpath(vcs_cache_dir)

        if os.path.isdir( vcs_cache_dir ):
            remote_info = self.vcsGetRepo( config, vcs_cache_dir )

            if remote_info != vcsrepo :
                self._warn('removing Hg cache on remote URL mismatch: {0} != {1}'
                    .format(remote_info, vcsrepo))
                self._rmTree( vcs_cache_dir )
            else :
                self._callExternal( [ hgBin, '--cwd', vcs_cache_dir, 'pull' ] )

        if not os.path.isdir( vcs_cache_dir ):
            self._callExternal( [ hgBin, 'clone', vcsrepo, vcs_cache_dir ] )
            
        return vcs_cache_dir

        
    def vcsGetRefRevision( self, config, vcs_cache_dir, branch ) :
        vcs_cache_dir = self._hgCache( config, vcs_cache_dir )
        
        hgBin = config['env']['hgBin']
        
        res = self._callExternal( [
            hgBin, '--repository', vcs_cache_dir, 'branches'
        ], verbose=False ).strip().split("\n")
        
        for r in res:
            r = r.split()
            
            if r and r[0] == branch:
                return r[1]
            
        self._errorExit( "Uknown Hg ref: {0}".format( branch ) )

    def vcsListTags( self, config, vcs_cache_dir, tag_hint ) :
        vcs_cache_dir = self._hgCache( config, vcs_cache_dir )
        
        hgBin = config['env']['hgBin']
        
        res = self._callExternal( [
            hgBin, '--repository', vcs_cache_dir, 'tags'
        ], verbose=False ).strip().split("\n")
        
        res = [ v and v.split()[0] or '' for v in res ]
        res = list(filter(None, res))
        del res[res.index('tip')]
        return res
    
    def vcsListBranches( self, config, vcs_cache_dir, branch_hint ) :
        vcs_cache_dir = self._hgCache( config, vcs_cache_dir )
        
        hgBin = config['env']['hgBin']
        
        res = self._callExternal( [
            hgBin, '--repository', vcs_cache_dir, 'branches'
        ], verbose=False ).strip().split("\n")
        
        res = [ v and v.split()[0] or '' for v in res ]
        res = list(filter(None, res))
        return res

    def vcsExport( self, config, vcs_cache_dir, vcs_ref, dst_path ) :
        vcs_cache_dir = self._hgCache( config, vcs_cache_dir )
        
        self._callExternal([
            config['env']['hgBin'],
            '--repository', vcs_cache_dir,
            'archive',
            '--rev', vcs_ref,
            '--type', 'files',
            #'--exclude', '.hg*',
            dst_path
        ])
            
        for f in ['.hgtags', '.hgignore', '.hg_archival.txt']:
            f = os.path.join(dst_path, f)

            if os.path.exists(f) :
                os.remove(f)
        
    
    def vcsBranch( self, config, vcs_ref ):
        hgBin = config['env']['hgBin']
        self._callExternal( [ hgBin, 'branch', vcs_ref ] )
        # We do not use bookmarks, so the only way to actually create the branch.
        self.vcsCommit(config, "CID new branch " + vcs_ref, [])
        self.vcsPush(config, ['--new-branch', vcs_ref])

    def vcsMerge( self, config, vcs_ref, cleanup ):
        curr_ref = self._getCurrentBranch(config)
        
        hgBin = config['env']['hgBin']
        
        try:
            self._callExternal( [ hgBin, 'merge', '--tool', 'internal:merge', vcs_ref ] )
        except subprocess.CalledProcessError:
            if cleanup:
                self.vcsRevert(config)
                self._errorExit('Merge failed, aborted.')
            self._errorExit('Merge failed, left as-is.')
        
        self.vcsCommit(config, "CID merged " + vcs_ref, [])
        self.vcsPush(config, [curr_ref])

    def vcsDelete( self, config, vcs_cache_dir, vcs_ref ):
        vcs_cache_dir = self._hgCache( config, vcs_cache_dir )
        hgBin = config['env']['hgBin']

        old_cwd = os.getcwd()
        os.chdir(vcs_cache_dir)
        
        self._callExternal( [ hgBin, 'checkout', '--check', vcs_ref ] + self._hgCheckoutTool(config) )
        self.vcsCommit(config, "CID branch delete", ['--close-branch'])
        self.vcsPush(config, [vcs_ref])
        os.chdir(old_cwd)

    def vcsRevert( self, config):
        hgBin = config['env']['hgBin']
        self._callExternal( [
            hgBin,
            'update', '--clean',
        ] + self._hgCheckoutTool(config) )
        self._callExternal( [
            hgBin,
            '--config', 'extensions.purge=',
            'purge', '-I', '*.orig', '-I', '**/*.orig', '--all'
        ])

    def vcsIsMerged( self, config, vcs_ref ):
        res = self._callExternal( [
            config['env']['hgBin'],
            'merge', '--preview',
            '--tool=internal:merge',
            vcs_ref
        ], verbose=False ).strip()
        return res == ''
    




import os

from .subtool import SubTool

__all__ = ['VcsTool']

class VcsTool( SubTool ):
    def autoDetect(self, config):
        return self._autoDetectVCS( config, self.autoDetectFiles() )

    def vcsGetRepo( self, config, wc_dir=None ):
        raise NotImplementedError( self._name )
    
    def vcsCheckout( self, config, branch ):
        raise NotImplementedError( self._name )
    
    def vcsCommit( self, config, message, files ):
        raise NotImplementedError( self._name )
    
    def vcsTag( self, config, tag, message ):
        raise NotImplementedError( self._name )
    
    def vcsPush( self, config, refs ):
        raise NotImplementedError( self._name )
    
    def vcsGetRevision( self, config ) :
        raise NotImplementedError( self._name )
    
    def vcsGetRefRevision( self, config, vcs_cache_dir, branch ) :
        raise NotImplementedError( self._name )

    def vcsListTags( self, config, vcs_cache_dir, tag_hint ) :
        raise NotImplementedError( self._name )

    def vcsListBranches( self, config, vcs_cache_dir, branch_hint ) :
        raise NotImplementedError( self._name )

    def vcsExport( self, config, vcs_cache_dir, vcs_ref, dst_path ) :
        raise NotImplementedError( self._name )
    
    def vcsBranch( self, config, vcs_ref ):
        raise NotImplementedError( self._name )

    def vcsMerge( self, config, vcs_ref, cleanup ):
        raise NotImplementedError( self._name )

    def vcsDelete( self, config, vcs_cache_dir, vcs_ref ):
        raise NotImplementedError( self._name )
    
    def vcsRevert( self, config ):
        raise NotImplementedError( self._name )
    
    def vcsIsMerged( self, config, vcs_ref ):
        raise NotImplementedError( self._name )

    def _autoDetectVCS( self, config, vcsDir ) :
        if config.get( 'vcs', None ) == self._name :
            return True
        
        if vcsDir in config['projectRootSet']:
            if config.get( 'vcs', self._name ) != self._name:
                self._errorExit( 'Another VCS type {0} has been already detected!'.format(config['vcs']) )
            config['vcs'] = self._name
            return True
        
        return False




from __future__ import print_function, absolute_import

import os

from .mixins.util import UtilMixIn
from .mixins.path import PathMixIn
from .mixins.package import PackageMixIn
from .mixins.version import VersionMixIn

__all__ = ['SubTool']

class SubTool( PathMixIn, PackageMixIn, UtilMixIn, VersionMixIn, object ):
    SYSTEM_VER = 'system'
    
    def __init__( self, name ) :
        self._name = name
        self._have_tool = False

    def getDeps( self ) :
        return []
    
    def envDeps( self, env ):
        pass
    
    def getPostDeps( self ) :
        return []
    
    def getOrder( self ):
        return 0
    
    def _installTool( self, env ):
        self._errorExit( 'Tool "{0}" must be manually installed'.format(self._name) )
    
    def envNames( self ) :
        return [ self._name + 'Bin' ]
    
    def importEnv( self, env ):
        environ = os.environ

        for name in self.envNames():
            val = environ.get(name, None)
            if val is not None:
                env[name] = val
                
    def exportEnv( self, env, dst):
        for name in self.envNames():
            if name in env:
                dst[name] = env[name]
    
    def initEnv( self, env, bin_name=None ) :
        name = self._name
        bin_env = name + 'Bin'

        if not env.get(bin_env, None) :
            if bin_name is None:
                bin_name = name

            tool_path = self._which( bin_name )
            if tool_path :
                env[ bin_env ] = tool_path.strip()
                self._have_tool = True
        else :
            self._have_tool = True
    
    def autoDetect( self, config ) :
        files = self.autoDetectFiles()
        
        if files:
            return self._autoDetectByCfg(config, files)

        return False
    
    def requireInstalled( self, env ) :
        self.importEnv( env )
        self.initEnv( env )

        if not self._have_tool:
            if self._isExternalToolsSetup( env ):
                self._errorExit( 'Tool "{0}" must be installed externally (env config)'.format(self._name) )
            else :
                self._warn('Auto-installing required "{0}" tool'.format(self._name))
                self._installTool( env )

            self.initEnv( env )
            
            if not self._have_tool:
                self._errorExit( 'Failed to install "{0}"'.format(self._name) )

    def isInstalled( self, env ):
        self.initEnv( env )
        return self._have_tool

    def updateTool( self, env ):
        self.requireInstalled( env )
        
    def uninstallTool( self, env ):
        self._have_tool = False
        self._errorExit( 'Tool "{0}" must be uninstalled externally'.format(self._name) )

    def loadConfig( self, config ) :
        pass
    
    def updateProjectConfig( self, config, updates ) :
        """
updates = {
    name : '...',
    version : '...',
}
@return a list of files to be committed
"""
        return []
    
    def onExec( self, env, args ):
        bin = env.get(self._name + 'Bin', None)

        if bin :
            self._callInteractive([bin] + args)
        else :
            self._errorExit( 'Exec command has not been implemented for "{0}"'.format(self._name) )





import re

from ..buildtool import BuildTool
from ..runenvtool import RunEnvTool

class dockerTool( BuildTool, RunEnvTool ):
    """Docker - Build, Ship, and Run Any App, Anywhere.
    
Home: https://www.docker.com/

Experimental support.

Docker CE support is added for CentOS, Fedora, Debian and Ubuntu.
For other systems, "docker" or "docker-engine" packages is tried to be installed.

Docker EE or other installation methods are out of scope for now.
"""
    def autoDetectFiles( self ) :
        return 'Dockerfile'
    
    def getDeps( self ):
        return ['curl']

    def getOrder( self ):
        return 10
    
    def envNames(self):
        return ['dockerBin', 'dockerVer', 'dockerRepos']
    
    def _installTool( self, env ):
        repo = env.get('dockerRepos', 'https://download.docker.com')
        
        if self._isCentOS():
            self._addYumRepo('docker', repo + '/linux/centos/docker-ce.repo')
            
        elif self._isFedora():
            self._addYumRepo('docker', repo + '/linux/fedora/docker-ce.repo')
            
        elif self._isDebian():
            gpg = self._callExternal([ env['curlBin'], '-fsSL', repo+'/linux/debian/gpg'])
            self._addAptRepo(
                'docker',
                'deb [arch=amd64] {0}/linux/debian $codename$ stable'.format(repo),
                gpg
            )

        elif self._isUbuntu():
            gpg = self._callExternal([ env['curlBin'], '-fsSL', repo+'/linux/ubuntu/gpg'])
            self._addAptRepo(
                'docker',
                'deb [arch=amd64] {0}/linux/ubuntu $codename$ stable'.format(repo),
                gpg,
                codename_map = {
                    'zesty': 'yakkety',
                },
                repo_base = repo,
            )
                
        elif self._isOracleLinux() or self._isRHEL():
            self._addYumRepo('docker', repo + '/linux/centos/docker-ce.repo')
            
        #elif self.isOpenSUSE() or self.isSLES():
        #    virt_repo = 'https://download.opensuse.org/repositories/Virtualization'
        #    
        #    with open('/etc/os-release', 'r') as rf:
        #        releasever = re.search('VERSION="([0-9.]+)"', rf.read()).group(1)
        #
        #    
        #    if self.isOpenSUSE():
        #        virt_repo += '/openSUSE_Leap_'+releasever
        #    else:
        #        virt_repo += '/SLE_'+releasever
        #    
        #    virt_gpg = self._callExternal([ env['curlBin'], '-fsSL', virt_repo+'/repodata/repomd.xml.key'])
        #    self._addZypperRepo('Virtualization', virt_repo+'/Virtualization.repo', virt_gpg)
        #    
        #    gpg = self._callExternal([ env['curlBin'], '-fsSL', repo+'/linux/centos/gpg'])
        #    self._addZypperRepo('docker', repo + '/linux/centos/7/x86_64/stable/', gpg, yum=True)
            
        else:
            self._requireYumEPEL()
            self._requirePackages(['docker'])
            self._requirePackages(['docker-engine'])
            self._requireEmerge(['app-emulation/docker'])
            self._requirePacman(['docker'])
            
            self._trySudoCall(
                ['/bin/systemctl', 'start', 'docker'],
                errmsg = 'you may need to start Docker manually !'
            )
            
            return
            
        ver = env.get('dockerVer', None)
        
        if ver:
            self._requirePackages(['docker-ce-'+ver])
        else :
            self._requirePackages(['docker-ce'])
            
        self._trySudoCall(
            ['/bin/systemctl', 'start', 'docker'],
            errmsg = 'you may need to start Docker manually !'
        )

    def onBuild( self, config ):
        cmd = [ config['env']['dockerBin'], 'build', '.' ]

        if self._haveGroup('docker'):
            self._callExternal( cmd )
        else:
            self._callExternal( ['sudo'] + cmd )





import os, subprocess

from ..runenvtool import RunEnvTool
from .bashtoolmixin import BashToolMixIn

class gvmTool( BashToolMixIn, RunEnvTool ):
    """Go Version Manager.

Home: https://github.com/moovweb/gvm
"""
    GVM_DIR_DEFAULT = os.path.join(os.environ['HOME'], '.gvm')
    GVM_VERSION_DEFAULT = 'master'
    GVM_INSTALLER_DEFAULT = 'https://raw.githubusercontent.com/moovweb/gvm/master/binscripts/gvm-installer'

    def getDeps( self ) :
        return [ 'bash', 'git', 'hg', 'curl', 'make', 'binutils' ]

    def _installTool( self, env ):
        self._requireDeb(['bison', 'gcc', 'build-essential'])
        self._requireRpm(['bison', 'gcc', 'glibc-devel'])
        self._requireEmergeDepsOnly(['dev-lang/go'])
        self._requirePacman(['bison', 'gcc', 'glibc',])

        self._callBash( env,
              '{0} < <({1} -s -S -L {2})'
               .format(env['bashBin'], env['curlBin'], env['gvmInstaller']),
               suppress_fail=True # error when Go is not yet installed
        )
            
    def updateTool( self, env ):
        self._installTool( env )
    
    def uninstallTool( self, env ):
        gvm_dir = env['gvmDir']
        
        if os.path.exists(gvm_dir):
            self._rmTree(gvm_dir)
            
        self._have_tool = False
            
    def envNames( self ) :
        return ['gvmDir', 'gvmInstaller']
    
    def initEnv( self, env ) :
        gvm_dir = env.setdefault('gvmDir', self.GVM_DIR_DEFAULT)
        os.environ['GVM_DEST'] = os.path.dirname(gvm_dir)
        os.environ['GVM_NAME'] = os.path.basename(gvm_dir)
        os.environ['GVM_NO_UPDATE_PROFILE'] = '1'
        
        env.setdefault('gvmVer', self.GVM_VERSION_DEFAULT)
        env.setdefault('gvmInstaller', self.GVM_INSTALLER_DEFAULT)
        
        env_init = os.path.join(gvm_dir, 'scripts', 'gvm')
        env['gvmInit'] = env_init
        
        self._have_tool = os.path.exists(env_init)
        
    def onExec( self, env, args ):
        self._callBashInteractive(env,
                '. {0} && gvm {1}'
                .format(env['gvmInit'], subprocess.list2cmdline(args))
        )




import os

from ..buildtool import BuildTool

class pipTool( BuildTool ):
    """The PyPA recommended tool for installing Python packages.
    
Home: https://pypi.python.org/pypi/pip
"""    
    REQUIREMENTS_FILE = 'requirements.txt'

    def autoDetectFiles( self ) :
        return self.REQUIREMENTS_FILE
    
    def getDeps( self ) :
        return [ 'python', 'virtualenv' ]

    def envNames( self ) :
        return ['pipBin', 'pipVer']

    def _installTool( self, env ):
        if os.path.exists( env['pipBin'] ):
            self.updateTool( env )
        else:
            self._callExternal([
                os.path.join( env['virtualenvDir'], 'bin', 'easy_install' ), 'pip'
            ])

    def updateTool( self, env ):
        self._callExternal([
            env['pipBin'], 'install', '-q',
            '--upgrade',
            'pip>={0}'.format(env['pipVer']),
        ])
        
    def uninstallTool( self, env ):
        pass
    
    def initEnv( self, env ) :
        pipBin = os.path.join( env['virtualenvDir'], 'bin', 'pip' )
        pipBin = env.setdefault( 'pipBin', pipBin )
        pipVer = env.setdefault( 'pipVer', '9.0.1' )
        
        if os.path.exists( pipBin ):
            pipFactVer = self._callExternal([ pipBin, '--version' ], verbose=False)
            pipFactVer = [int(v) for v in pipFactVer.split(' ')[1].split('.')]
            pipNeedVer = [int(v) for v in pipVer.split('.')]
            
            self._have_tool = pipNeedVer <= pipFactVer
    
    def onPrepare( self, config ):
        if os.path.exists(self.REQUIREMENTS_FILE):
            self._callExternal( [ config['env']['pipBin'], 'install', '-r', self.REQUIREMENTS_FILE ] )




import os
import re

from ..rmstool import RmsTool

class scpTool( RmsTool ):
    """secure copy."""
    
    REMOTE_PATTERN = '^([a-zA-Z][a-zA-Z0-9_]+@[a-zA-Z][a-zA-Z0-9_]+)(:(.+))?$'
    REMOTE_GRP_USER_HOST = 1
    REMOTE_GRP_PATH = 3
    
    def getDeps( self ) :
        return [ 'ssh' ]
    
    def rmsPromote( self, config, package, rms_pool ) :
        scpBin = config['env']['scpBin']
        rms_repo = config['rmsRepo']

        package_basename = os.path.basename( package )
        dst = os.path.join( rms_repo, rms_pool, package_basename )
        
        if os.path.exists( package ) :
            self._callExternal( [ scpBin, '-Bq', package, dst ] )
        else :
            src = os.path.join( rms_repo, package )

            self._callExternal( [ scpBin, '-Bq', src, package_basename ] )
            self.rmsVerifyHash( config, package_basename )

            # Note: promotion must no re-upload an artifact to avoid
            # risk of modifications
            self._callExternal( [ scpBin, '-Bq', src, dst ] )

    def rmsGetList( self, config, rms_pool, package_hint ):
        env = config['env']
        rms_repo = config['rmsRepo']
        result = re.match( self.REMOTE_PATTERN, rms_repo )
        
        if result:
            user_host = result.group( self.REMOTE_GRP_USER_HOST )
            path = os.path.join( result.group( self.REMOTE_GRP_PATH ), rms_pool )
            cmd = 'ls {0}'.format( path )
            ret = self._callExternal( [ env['sshBin'], '-t', user_host, '--', cmd ] ).split("\n")
        else:
            path = os.path.join( rms_repo, rms_pool )  
            ret = os.listdir(path)
    
        return ret
    
    def rmsRetrieve( self, config, rms_pool, package ):
        scpBin = config['env']['scpBin']
        package_basename = os.path.basename( package )
        src = os.path.join( config['rmsRepo'], rms_pool, package_basename )
        self._callExternal( [ scpBin, '-Bq', src, package_basename ] )
    



from ..rmstool import RmsTool
from .piptoolmixin import PipToolMixIn

class twineTool( PipToolMixIn, RmsTool ):
    """Collection of utilities for interacting with PyPI
Home: https://pypi.python.org/pypi/twine

Not a fully supported RMS. Supports only uploading of local packages.

Note: rmsRepo is ignored and rmsPool is actual repo URL or uses ~/.pypirc entry
"""
    def getDeps( self ) :
        return super(twineTool, self).getDeps() + [ 'gpg' ]
    
    def envNames(self):
        return [
            'twineBin',
            'twineIdentity',
            'twineUser',
            'twinePassword',
        ]
        
    def rmsPromoteMany( self, config, packages, rms_pool ):
        env = config['env']
        
        identity = env.get('twineIdentity', None)
        user = env.get('twineUser', None)
        passwd = env.get('twinePassword', None)
        
        cmd = [ env['twineBin'], 'upload' ]
        cmd += ['--repository', rms_pool]
        
        if identity:
            cmd += ['--sign', '--identity', identity]
        elif self._getTune(config, 'requireSign', True):
            self._errorExit('Twine config requires GPG signature. Please set "twineIdentity"')
            
        if user:
            cmd += ['--username', user]
        
        if passwd:
            cmd += ['--password', passwd]

        cmd.append('--skip-existing')
        
        cmd += packages
        
        self._callInteractive(cmd, replace=False)
    
    def rmsPromote( self, config, package, rms_pool ):
        self.rmsPromoteMany( config, [package], rms_pool )



from ..buildtool import BuildTool
from ..testtool import TestTool
from .sdkmantoolmixin import SdkmanToolMixIn

class sbtTool( SdkmanToolMixIn, BuildTool, TestTool ):
    """The interactive build tool (Scala).
    
Home: http://www.scala-sbt.org/

Installed via SDKMan!

First run of SBT may consume a lot of time on post-installation steps.

Build targets:
    prepare -> clean
    build -> compile
    package -> package
    check -> test
Override targets with .config.toolTune.

Requires Java >= 8.
"""
    _MIN_JAVA = '8'
    
    def autoDetectFiles( self ) :
        return 'build.sbt'
    
    def onPrepare( self, config ):
        target = self._getTune(config, 'prepare', 'clean')
        self._callExternal( [ config['env']['sbtBin'], target ] )

    def onBuild( self, config ):
        target = self._getTune(config, 'build', 'compile')
        self._callExternal( [ config['env']['sbtBin'], target ] )

    def onPackage( self, config ):
        target = self._getTune(config, 'package', 'package')
        self._callExternal( [ config['env']['sbtBin'], target ] )
        self._addPackageFiles(config, 'target/scala-*/*.jar')
            
    def onRunDev( self, config ):
        target = self._getTune(config, 'run', 'check')
        self._callExternal( [ config['env']['sbtBin'], target ] )
            
    def onCheck( self, config ):
        target = self._getTune(config, 'check', 'test')
        self._callExternal( [ config['env']['sbtBin'], target ] )




from ..buildtool import BuildTool
from .npmtoolmixin import NpmToolMixIn

class gulpTool( NpmToolMixIn, BuildTool ):
    """Automate and enhance your workflow (Node.js).

Home: http://gulpjs.com/    
"""    
    def autoDetectFiles( self ) :
        return 'gulpfile.js'
    
    def onBuild( self, config ):
        self._callExternal( [ config['env']['gulpBin'] ] )






from ..buildtool import BuildTool
from .piptoolmixin import PipToolMixIn


class cidTool( PipToolMixIn, BuildTool ):
    "Noop FutoIn-CID - a workaround to allow CID use from virtualenv"

    def _pipName( self ):
        return 'futoin-cid'




__version__ = '0.7.3'

from .cli import run
from .subtool import SubTool
from .buildtool import BuildTool
from .rmstool import RmsTool
from .runenvtool import RunEnvTool
from .vcstool import VcsTool




import os, glob

from ..runtimetool import RuntimeTool
from .javatoolmixin import JavaToolMixIn

class javaTool( RuntimeTool, JavaToolMixIn ):
    """Java Runtime Environment.
    
Home: http://openjdk.java.net/

Due to issues with Oracle's licensing, cid
supports only automatic installation of OpenJDK.

javaVer supports:
- only one digit like 6, 7, 8, 9 (Zulu JDK (http://zulu.org)
- use javaBin for custom JRE
"""
    def envNames( self ) :
        return ['javaBin', 'javaVer']
    
    def getVersionParts( self ):
        return 1
    
    def _installTool( self, env ):
        if env.get('javaBin', None):
            return

        ver = env['javaVer']
        
        if self._isMacOS():
            if ver == '8':
                self._requireDmg('https://cdn.azul.com/zulu/bin/zulu8.20.0.5-jdk8.0.121-macosx_x64.dmg')
            elif ver == '7':
                self._requireDmg('https://cdn.azul.com/zulu/bin/zulu7.17.0.5-jdk7.0.131-macosx_x64.dmg')
            return
        
        self._addAptRepo('zulu', 'deb http://repos.azulsystems.com/debian stable main', self._ZULU_GPG_KEY)
        self._addYumRepo('zulu', 'http://repos.azulsystems.com/rhel/zulu.repo', self._ZULU_GPG_KEY, releasevermax=7)
        self._addZypperRepo('zulu', 'http://repos.azulsystems.com/sles/latest', self._ZULU_GPG_KEY)
        
        # leaving here for possible future use
        #self._requireDeb(['openjdk-{0}-jre-headless'.format(ver)])
        #self._requireYum(['java-1.{0}.0-openjdk'.format(ver)])
        #self._requireZypper(['java-1_{0}_0-openjdk'.format(ver)])
        self._requirePackages(['zulu-{0}'.format(ver)])
        
        self._requirePacman(['jre{0}-openjdk'.format(ver)])
        self._requireEmerge(['=dev-java/oracle-jre-bin-1.{0}*'.format(ver)])

    
    def uninstallTool( self, env ):
        pass

    def initEnv( self, env ) :
        if env.get('javaBin', None):
            bin_dir = os.path.dirname(env['javaBin'])
            java_home = os.path.dirname(env['javaBin'])
            
            os.environ['JAVA_HOME'] = java_home
            os.environ['JRE_HOME'] = java_home
            self._addBinPath(bin_dir, True)
            
            super(javaTool, self).initEnv( env, 'java' )
            return
        
        env.setdefault('javaVer', self._LATEST_JAVA)
        ver = env['javaVer']
        
        candidates = [
            # Zulu
            '/usr/lib/jvm/zulu-{0}*/jre/bin/java'.format(ver),
            # Debian / Ubuntu
            #"/usr/lib/jvm/java-{0}-openjdk*/jre/bin/java".format(ver),
            # RedHat
            #"/usr/lib/jvm/jre-1.{0}.0/bin/java".format(ver),
            # OpenSuse
            #"/usr/lib*/jvm/jre-1.{0}.0/bin/java".format(ver),
            # Default oracle
            #"/opt/jdk/jdk1.{0}*/bin/java".format(ver),
        ]

        if self._isGentoo() or self._isArchLinux():
            candidates += [
                "/usr/lib/jvm/java-{0}-openjdk*/jre/bin/java".format(ver),
            ]
        
        for c in candidates:
            bin_name = glob.glob(c)
            
            if bin_name:
                bin_name = bin_name[0]
                bin_dir = os.path.dirname(bin_name)
                java_home = os.path.dirname(bin_dir)
                
                env['javaBin'] = bin_name
                os.environ['JAVA_HOME'] = java_home
                os.environ['JRE_HOME'] = java_home
                self._addBinPath(bin_dir, True)
                self._have_tool = True
                break
                
    def onRun( self, config, file, args, tune ):
        env = config['env']
        self._callInteractive([
            env[self._name + 'Bin'], '-jar', file
        ] + args)
            
    _ZULU_GPG_KEY = '''
-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: GnuPG v1

mQINBFNgFa8BEADTL/REB10M+TfiZOtFHqL5LHKkzTMn/O2r5iIqXGhi6iwZazFs
9S5g1eU7WMen5Xp9AREs+OvaHx91onPZ7ZiP7VpZ6ZdwWrnVk1Y/HfI59tWxmNYW
DmKYBGMj4EUpFPSE9EnFj7dm1WdlCvpognCwZQl9D3BseGqN7OLHfwqqmOlbYN9h
HYkT+CaqOoWDIGMB3UkBlMr0GuujEP8N1gxg7EOcSCsZH5aKtXubdUlVSphfAAwD
z4MviB39J22sPBnKmaOT3TUTO5vGeKtC9BAvtgA82jY2TtCEjetnfK/qtzj/6j2N
xVUbHQydwNQVRU92A7334YvCbn3xUUNI0WOscdmfpgCU0Z9Gb2IqDb9cMjgUi8F6
MG/QY9/CZjX62XrHRPm3aXsCJOVh/PO1sl2A/rvv8AkpJKYyhm6T8OBFptCsA3V4
Oic7ZyYhqV0u2r4NON+1MoUeuuoeY2tIrbRxe3ffVOxPzrESzSbc8LC2tYaP+wGd
W0f57/CoDkUzlvpReCUI1Bv5zP4/jhC63Rh6lffvSf2tQLwOsf5ivPhUtwUfOQjg
v9P8Wc8K7XZpSOMnDZuDe9wuvB/DiH/P5yiTs2RGsbDdRh5iPfwbtf2+IX6h2lNZ
XiDKt9Gc26uzeJRx/c7+sLunxq6DLIYvrsEipVI9frHIHV6fFTmqMJY6SwARAQAB
tEdBenVsIFN5c3RlbXMsIEluYy4gKFBhY2thZ2Ugc2lnbmluZyBrZXkuKSA8cGtp
LXNpZ25pbmdAYXp1bHN5c3RlbXMuY29tPokCOAQTAQIAIgUCU2AVrwIbAwYLCQgH
AwIGFQgCCQoLBBYCAwECHgECF4AACgkQsZmDYSGb2cnJ8xAAz1V1PJnfOyaRIP2N
Ho2uRwGdPsA4eFMXb4Z08eGjDMD3b9WW3D0XnCLbJpaZ6klz0W0s2tcYSneTBaSs
RAqxgJgBZ5ZMXtrrHld/5qFoBbStLZLefmcPhnfvamwHDCTLUex8NIAI1u3e9Rhb
5fbH+gpuYpwHX7hz0FOfpn1sxR03UyxU+ey4AdKe9LG3TJVnB0WcgxpobpbqweLH
yzcEQCNoFV3r1rlE13Y0aE31/9apoEwiYvqAzEmE38TukDLl/Qg8rkR1t0/lok2P
G6pWqdN7pmoUovBTvDi5YOthcjZcdOTXXn2Yw4RZVF9uhRsVfku1Eg25SnOje3uY
smtQLME4eESbePdjyV/okCIle66uHZse+7gNyNmWpf01hM+VmAySIAyKa0Ku8AXZ
MydEcJTebrNfW9uMLsBx3Ts7z/CBfRng6F8louJGlZtlSwddTkZVcb26T20xeo0a
ZvdFXM2djTi/a5nbBoZQL85AEeV7HaphFLdPrgmMtS8sSZUEVvdaxp7WJsVuF9cO
Nxsvx40OYTvfco0W41Lm8/sEuQ7YueEVpZxiv5kX56GTU9vXaOOi+8Z7Ee2w6Adz
4hrGZkzztggs4tM9geNYnd0XCdZ/ICAskKJABg7biDD1PhEBrqCIqSE3U497vibQ
Mpkkl/Zpp0BirhGWNyTg8K4JrsQ=
=d320
-----END PGP PUBLIC KEY BLOCK-----
'''
            




from __future__ import print_function, absolute_import

import os, tempfile, subprocess, platform, glob

try:
    import urllib.request as urllib
except ImportError:
    import urllib2 as urllib

class PackageMixIn( object ):
    def _isCentOS( self ):
        return platform.linux_distribution()[0].startswith('CentOS')
    
    def _isFedora( self ):
        return os.path.exists('/etc/fedora-release')
    
    def _isGentoo( self ):
        return os.path.exists('/etc/gentoo-release')

    def _isArchLinux( self ):
        # There are cases when platform.linux_distribution() is empty on Arch
        return (os.path.exists('/etc/arch-release') or
                platform.linux_distribution()[0].startswith('arch'))
    
    def _isDebian( self ):
        return platform.linux_distribution()[0].startswith('debian')

    def _isUbuntu( self ):
        return platform.linux_distribution()[0].startswith('Ubuntu')
    
    def _isOracleLinux( self ):
        return platform.linux_distribution()[0].startswith('Oracle Linux')

    def _isRHEL( self ):
        return platform.linux_distribution()[0].startswith('Red Hat Enterprise Linux')
    
    def isOpenSUSE( self ):
        return platform.linux_distribution()[0].startswith('openSUSE')

    def isSLES( self ):
        return platform.linux_distribution()[0].startswith('SUSE Linux Enterprise')
    
    def _isMacOS( self ):
        return platform.system() == 'Darwin'

    def _requireDeb(self, packages):
        apt_get = self._which('apt-get')
        
        if apt_get:
            if not isinstance(packages, list):
                packages = [packages]
            
            self._trySudoCall(
                [apt_get, 'install', '-y'] + packages,
                errmsg = 'you may need to install the packages manually !'
            )

    def _requireYum(self, packages):
        yum = self._which('dnf')
        
        if not yum:
            yum = self._which('yum')
        
        if yum:
            if not isinstance(packages, list):
                packages = [packages]
            
            self._trySudoCall(
                [yum, 'install', '-y'] + packages,
                errmsg = 'you may need to install the packages manually !'
            )

    def _requireZypper(self, packages):        
        zypper = self._which('zypper')

        if zypper:
            if not isinstance(packages, list):
                packages = [packages]
            
            self._trySudoCall(
                [zypper, 'install', '-y'] + packages,
                errmsg='you may need to install the packages manually !'
            )
            
    def _requireRpm(self, packages):
        self._requireYum(packages)
        self._requireZypper(packages)
    
    def _requirePackages(self, packages):
        self._requireDeb(packages)
        self._requireRpm(packages)
        
    def _requireEmerge(self, packages):
        emerge = self._which('emerge')

        if emerge:
            if not isinstance(packages, list):
                packages = [packages]
            
            self._trySudoCall(
                [emerge] + packages,
                errmsg='you may need to install the build deps manually !'
            )

    def _requireEmergeDepsOnly(self, packages):
        if not isinstance(packages, list):
            packages = [packages]
        
        self._requireEmerge(['--onlydeps'] + packages)
        
    def _requirePacman(self, packages):
        pacman = self._which('pacman')

        if pacman:
            if not isinstance(packages, list):
                packages = [packages]
            
            self._trySudoCall(
                [pacman, '-S', '--noconfirm', '--needed'] + packages,
                errmsg='you may need to install the build deps manually !'
            )
        
    def _addAptRepo(self, name, entry, gpg_key=None, codename_map=None, repo_base=None):
        self._requireDeb([
            'software-properties-common',
            'apt-transport-https',
            'ca-certificates',
            'lsb-release',
        ])
        apt_add_repository = self._which('apt-add-repository')
        
        if not apt_add_repository:
            return
        
        if gpg_key:
            (fd, tf) = tempfile.mkstemp('cidgpg', text=True)
            try:
                os.write(fd, gpg_key.encode(encoding='UTF-8'))
            except:
                os.write(fd, gpg_key)
            os.close(fd)

            self._trySudoCall(
                ['apt-key', 'add', tf],
                errmsg = 'you may need to import the PGP key manually!'
            )
            
            os.remove(tf)
            
        codename = subprocess.check_output( ['lsb_release', '-cs'] )
        try: codename = str(codename, 'utf8')
        except: pass
        codename = codename.strip()
        
        if codename_map:
            try:
                repo_info = urllib.urlopen('{0}/{1}'.format(repo_base, codename)).read()
            except:
                fallback_codename = codename_map.get(codename, codename)
                self._warn('Fallback to codename: {0}'.format(fallback_codename))
                codename = fallback_codename

        entry = entry.replace('$codename$', codename)
            
        self._trySudoCall(
            [apt_add_repository, '--yes', entry],
            errmsg = 'you may need to add the repo manually!'
        )
        
        self._trySudoCall(
            ['apt-get', 'update'],
            errmsg = 'you may need to update APT cache manually!'
        )
            
    def _addRpmKey(self, gpg_key):
        if not gpg_key:
            return
        
        rpm = self._which('rpm')
        
        if not rpm:
            return
            
        (fd, tf) = tempfile.mkstemp('cidgpg', text=True)
        os.write(fd, gpg_key.encode(encoding='UTF-8'))
        os.close(fd)
            
        self._trySudoCall(
            [rpm, '--import', tf],
            errmsg = 'you may need to import the PGP key manually!'
        )
        
        os.remove(tf)
            
    
    def _addYumRepo(self, name, url, gpg_key=None, releasevermax=None):
        self._addRpmKey(gpg_key)

        dnf = self._which('dnf')
        yum = self._which('yum')
        
        if dnf:
            self._requireYum(['dnf-plugins-core'])
            repo_file = None
            
            if releasevermax is not None:
                dump = self._callExternal([dnf, 'config-manager', '--dump'], verbose=False)
                for l in dump.split("\n"):
                    l = l.split(' = ')
                    
                    if l[0] == 'releasever':
                        if int(l[1]) > releasevermax:
                            repo_info = urllib.urlopen(url).read()
                            
                            try: repo_info = str(repo_info, 'utf8')
                            except: pass
                        
                            repo_info = repo_info.replace('$releasever', str(releasevermax))
                        
                            tmp_dir = tempfile.mkdtemp('cidrepo')
                            repo_file = url.split('/')[-1]
                            repo_file = os.path.join(tmp_dir, repo_file)
                            
                            with open(repo_file, 'w') as f:
                                f.write(repo_info)
                            
                            url = repo_file
                        break

            self._trySudoCall(
                [dnf, 'config-manager', '--add-repo', url],
                errmsg = 'you may need to add the repo manually!'
            )
            
            if repo_file:
                os.remove(repo_file)
            
            
        elif yum:
            self._requireYum(['yum-utils'])
            yumcfgmgr = self._which('yum-config-manager')
            self._trySudoCall(
                [yumcfgmgr, '--add-repo', url],
                errmsg = 'you may need to add the repo manually!'
            )

    def _addZypperRepo(self, name, url, gpg_key=None, yum=False):
        self._addRpmKey(gpg_key)
        
        zypper = self._which('zypper')
        
        if zypper:
            if yum:
                cmd = [zypper, 'addrepo', '-t', 'YUM', url, name]
            else:
                cmd = [zypper, 'addrepo', url, name]
            
            self._trySudoCall(
                cmd,
                errmsg = 'you may need to add the repo manually!'
            )
            
    def _requireYumEPEL(self):
        if self._isOracleLinux() or self._isRHEL():
            ver = platform.linux_distribution()[1].split('.')[0]
            self._requireYum(['https://dl.fedoraproject.org/pub/epel/epel-release-latest-{0}.noarch.rpm'.format(ver)])
        else:
            self._requireYum(['epel-release'])
            
    def _yumEnable(self, repo):
        self._requireYum(['yum-utils'])

        yumcfgmgr = self._which('yum-config-manager')

        self._trySudoCall(
            [yumcfgmgr, '--enable', repo],
            errmsg='You may need to enable the repo manually'
        )
        
    def _isSCLSupported(self):
        "Check if Software Collections are supported"
        return (
            self._isCentOS() or
            self._isRHEL() or
            self._isOracleLinux()
        )
        
    def _requireSCL(self):
        if self._isRHEL():
            self._yumEnable('rhel-server-rhscl-7-rpms')
        elif self._isCentOS():
            self._requireYum('centos-release-scl-rh')
        elif self._isOracleLinux():
            self._addYumRepo('public-yum-o17', 'http://yum.oracle.com/public-yum-ol7.repo')
            self._yumEnable('ol7_software_collections')
            self._yumEnable('ol7_latest')
            self._yumEnable('ol7_optional_latest')

        self._requireYum('scl-utils')

    def _requireHomebrew(self, packages):
        if not self._isMacOS():
            return
        
        if not isinstance(packages, list):
            packages = [packages]
        
        brew = self._which('brew')
        
        if not brew:
            self._callExternal(['bash', '-c', '/usr/bin/ruby -e "$(curl -fsSL https://raw.githubusercontent.com/Homebrew/install/master/install)"'])
            brew = self._which('brew')
            
        for package in packages:
            self._callExternal([brew, 'install', package])
            
    def _requireDmg(self, packages):
        if not self._isMacOS():
            return
        
        if not isinstance(packages, list):
            packages = [packages]

        curl = self._which('curl')
        hdiutil = self._which('hdiutil')
        installer = self._which('installer')
        volumes_dir = '/Volumes'
            
        for package in packages:
            base_name = package.split('/')[-1]
            local_name = os.path.join(os.environ['HOME'])
            self._callExternal([curl, '-o', base_name, package])
            
            volumes = set(os.listdir(volumes_dir))
            self._trySudoCall([hdiutil, 'attach', local_name])
            volume = (set(os.listdir(volumes_dir)) - volumes)[0]
            
            pkg = glob.glob(os.path.join(volumes_dir, volume, '*.pkg'))
            self._trySudoCall([installer, '-package', pkg, '-target', '/'])
            
            self._trySudoCall([hdiutil, 'dettach', local_name])



from ..runtimetool import RuntimeTool
from .bashtoolmixin import BashToolMixIn

class goTool( BashToolMixIn, RuntimeTool ):
    """The Go Programming Language
    
Home: https://golang.org/

All versions are installed through GVM.

Only binary releases of Golang are supported for installation
through CID, but you can install source releases through
"cid tool exec gvm -- install sourcetag".
"""
    def getDeps( self ) :
        return [ 'gvm', 'bash', 'binutils', 'gcc' ]
    
    def _installTool( self, env ):
        # in case GVM is already installed without these deps
        self._requireDeb(['bison', 'build-essential'])
        self._requireRpm(['bison', 'glibc-devel'])
        self._requireEmergeDepsOnly(['dev-lang/go'])
        self._requirePacman(['bison', 'glibc',])
        self._requireHomebrew('bison')

        self._callBash( env,
            'source {0} && gvm install {1} --binary'
            .format(env['gvmInit'], env['goVer'])
        )

    def updateTool( self, env ):
        self._installTool( env )
        
    def uninstallTool( self, env ):
        self._callBash( env,
            'source {0} && gvm uninstall {1}'
            .format(env['gvmInit'], env['goVer'])
        )
        self._have_tool = False

    def envNames( self ) :
        return ['goVer', 'goBin' ]

    def initEnv( self, env ) :
        if not env.get('goVer', None):
            try:
                env['goVer'] = self._callBash( env,
                    'source {0} && gvm listall | egrep "go[0-9]\.[0-9]+(\.[0-9]+)?$" | sort -rV'
                    .format(env['gvmInit']),
                    verbose=False
                ).split("\n")[0].strip()
            except:
                return
            
        ver = env['goVer']

        try:
            env_to_set = self._callBash( env,
                'source {0} && \
                gvm use {1} >/dev/null && \
                env | egrep -i "(gvm|golang)"'.format(env['gvmInit'], ver),
                verbose = False
            )
        except:
            return

        if env_to_set :
            self._updateEnvFromOutput(env_to_set)
            super(goTool, self).initEnv( env )
    
            
    def onRun( self, config, file, args, tune ):
        env = config['env']
        self._callInteractive([
            env[self._name + 'Bin'], 'run', file
        ] + args)
            
   



from setuptools import setup, find_packages
    
import os, sys

project_path = os.path.dirname( __file__ )
sys.path.insert( 0, project_path )
from futoin.cid import __version__ as version

with open(os.path.join(project_path, 'README.rst'), 'r') as f:
    long_description = f.read()

config = {
    'name': 'futoin-cid',
    'version': version,
    
    'description': 'FutoIn Continuous Integration & Delivery Tool',
    'long_description': long_description,

    'author': 'Andrey Galkin',
    'author_email': 'andrey@futoin.org',

    'url': 'https://github.com/futoin/cid-tool',
    'download_url': 'https://github.com/futoin/cid-tool/archive/master.zip',

    'install_requires': ['docopt'],
    'extras_require': {
        'test': ['nose'],
    },
    'python_requires': '>=2.7',
    'packages': find_packages(exclude=['bind', 'tests']),
    
    'entry_points': {
        'console_scripts': [
            'cid=futoin.cid.cli:run',
            'futoin-cid=futoin.cid.cli:run',
        ],
    },
    'classifiers': [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Software Development :: Build Tools',
        'Topic :: System :: Installation/Setup',
        'Programming Language :: C',
        'Programming Language :: C++',
        'Programming Language :: Java',
        'Programming Language :: JavaScript',
        'Programming Language :: Other',
        'Programming Language :: PHP',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Ruby',
        'Programming Language :: Unix Shell',
        'Environment :: Console',
    ],
    'keywords': 'php ruby node nodejs npm gem rvm nvm grunt gulp bower \
 puppet build deploy futoin cmake make gradle maven java composer bundler',
    'license': 'Apache 2.0',
}

setup(**config)




from .subtool import SubTool

__all__ = ['BuildTool']

class BuildTool( SubTool ):
    def onPrepare( self, config ):
        pass
    
    def onBuild( self, config ):
        pass
    
    def onPackage( self, config ):
        pass




from ..buildtool import BuildTool
from .npmtoolmixin import NpmToolMixIn

class gruntTool( NpmToolMixIn, BuildTool ):
    """Grunt: The JavaScript Task Runner.
    
Home: https://gruntjs.com/    
"""    
    def autoDetectFiles( self ) :
        return [ 'Gruntfile.js', 'Gruntfile.coffee' ]

    def _npmName(self):
        return 'grunt-cli'
    
    def onBuild( self, config ):
        gruntBin = config['env']['gruntBin']
        self._callExternal( [ gruntBin ] )
            



import os, subprocess

from ..runenvtool import RunEnvTool
from .bashtoolmixin import BashToolMixIn

class nvmTool( BashToolMixIn, RunEnvTool ):
    """Node Version Manager.

Home: https://github.com/creationix/nvm    
"""
    NVM_DIR_DEFAULT = os.path.join(os.environ['HOME'], '.nvm')
    NVM_LATEST = '$(git describe --abbrev=0 --tags --match "v[0-9]*")'

    def getDeps( self ) :
        return [ 'bash', 'git', 'curl' ]

    def _installTool( self, env ):
        nvm_dir = env['nvmDir']
        nvm_git = env.get('nvmGit', 'https://github.com/creationix/nvm.git')
        nvm_ver = env.get('nvmVer', self.NVM_LATEST) 

        self._callBash( env,
              'git clone {1} {0}; \
               cd {0} && git fetch && git reset --hard && git checkout {2}'
               .format(nvm_dir, nvm_git, nvm_ver) )
            
    def updateTool( self, env ):
        nvm_dir = env['nvmDir']
        nvm_ver = env.get('nvmVer', self.NVM_LATEST) 

        self._callBash( env,
              'cd {0} && git fetch && git reset --hard && git checkout {1}'
               .format(nvm_dir, nvm_ver) )
    
    def uninstallTool( self, env ):
        nvm_dir = env['nvmDir']
        
        if os.path.exists(nvm_dir):
            self._rmTree(nvm_dir)
            
        self._have_tool = False
            
    def envNames( self ) :
        return ['nvmDir', 'nvmGit', 'nvmVer']
    
    def initEnv( self, env ) :
        nvm_dir = env.setdefault('nvmDir', self.NVM_DIR_DEFAULT)
        env_init = os.path.join(nvm_dir, 'nvm.sh')
        env['nvmInit'] = env_init
        self._have_tool = os.path.exists(env_init)
        
    def onExec( self, env, args ):
        self._callBashInteractive(env,
                '. {0} && nvm {1}'
                .format(env['nvmInit'], subprocess.list2cmdline(args))
        )
